import numpy as np
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.postproc.dbs.units_dict import UnitsDict
from perturbopy.postproc.dbs.recip_pt_db import RecipPtDB
from perturbopy.postproc.utils.plot_tools import plot_dispersion, plot_recip_pt_labels


class Phdisp(CalcMode):
    """
    Class representation of a Perturbo phonon dispersion (phdisp) calculation.

   
    Attributes
    ----------
    qpt : RecipPtDB
       Database for the q-points used in the phdisp calculation, containing M points.
    
    phdisp : UnitsDict
       Database for the phonon energies computed by the phdisp calculation. The keys are
       the phonon mode, and the values are an array (of length M) containing the energies at each q-point
       with units phdisp.units

    """

    def __init__(self, pert_dict):
        """
        Constructor method

        Parameters
        ----------
        pert_dict : dict
           Dictionary containing the inputs and outputs from the phdisp calculation.

        """
        super().__init__(pert_dict)

        if self.calc_mode != 'phdisp':
            raise ValueError('Calculation mode for a Phdisp object should be "phdisp"')

        qpath_units = self._pert_dict['phdisp'].pop('q-path coordinate units')
        qpath = np.array(self._pert_dict['phdisp'].pop('q-path coordinates'))
        qpoint_units = self._pert_dict['phdisp'].pop('q-point coordinate units')
        qpoint = np.array(self._pert_dict['phdisp'].pop('q-point coordinates'))

        energies_dict = self._pert_dict['phdisp'].pop('phonon mode')
        num_modes = self._pert_dict['phdisp'].pop('number of modes')
        energy_units = self._pert_dict['phdisp'].pop('phdisp units')

        self.qpt = RecipPtDB.from_lattice(qpoint, qpoint_units, self.lat, self.recip_lat, qpath, qpath_units)
        self.phdisp = UnitsDict.from_dict(energies_dict, energy_units)

    def plot_phdisp(self, ax, show_qpoint_labels=True, c='k', ls='-', energy_window=None):
        """
        Method to plot the phonon dispersion.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
           Axis on which to plot the phonons.

        show_qpoint_labels : bool, optional
           If true, the q-point labels stored in the labels attribute will be shown on the plot.

        c : str, list
            Color(s) for phonon modes; see corresponding entry in plot_tools.plot_dispersion function

        ls : str, list
            Linestyle(s) for phonon modes; see corresponding entry in plot_tools.plot_dispersion function

        energy_window : tuple of float, optional
           The range of phonon energies to be shown on the y-axis.

        Returns
        -------
        ax: matplotlib.axes.Axes
           Axis with the plotted phonons.

        """
        ax = plot_dispersion(ax, self.qpt.path, self.phdisp, self.phdisp.units, c, ls, energy_window)

        if show_qpoint_labels:
            ax = plot_recip_pt_labels(ax, self.qpt.labels, self.qpt.points, self.qpt.path)

        return ax



import numpy as np
from perturbopy.postproc.dbs.units_dict import UnitsDict
from perturbopy.postproc.calc_modes.calc_mode import CalcMode


class Trans(CalcMode):
    """
    Class representation of a Perturbo imsigma calculation.

    Attributes
    ----------
    temper : UnitsDict
        Dictionary of temperatures used in each configuration. The keys give the configuration number,
        and the values are floats giving the temperature (with units temper.units)
    
    chem_pot : UnitsDict
        Dictionary of chemical potentials used in each configuration. The keys
        give the configuration number, and the values are floats giving the
        chemical potential (with units chem_pot.units)
    
    conc : UnitsDict
        Dictionary of carrier concentrations used in each configuration. The keys
        give the configuration number, and the values are floats giving the
        concentration (with units conc.units)
    
    cond : UnitsDict
        Dictionary of conductivity tensors computed for each configuration. The keys are the
        configuration number, and the values are 3x3 arrays giving the
        conductivity tensor (with units cond.units)
    
    mob : UnitsDict
        Dictionary of mobility tensors computed for each configuration. The keys give the
        configuration number, and the values are 3x3 arrays for the mobility tensor (with units mob.units)
    
    seebeck : UnitsDict
        Dictionary of Seebeck coefficient tensors computed for each configuration. The keys give the
        configuration number, and the values are 3x3 arrays giving theSeebeck coefficient tensor
        (with units seebeck.units). If not computed, this field will be None.
    
    thermal_cond : UnitsDict
        Dictionary of thermal conductivity tensors computed for each configuration. The keys give
        the configuration number, and the values are 3x3 arrays giving the thermal conductivity tensor
        (with units thermal_cond.units) If not computed, this field will be None.
    
    bfield : UnitsDict
        Dictionary of magnetic fields used in each configuration, if running a magnetic field transport
        calculation (if not, this field will be None). The keys are the configuration number, and the values
        are the 3-dimensional magnetic field (with units of bfield.units)
    
    cond_iter : UnitsDict
        Dictionary of dictionaries giving conductivity tensor computed at each iteration when solving
        the iterative BTE, if running an ITA calculation (if not, this field will be None).
        The keys of the top level dictionary are the configuration number, and the values are
        additional dictionaries with keys giving the iteration number and values giving the conductivity tensor
        at that iteration.

    """

    def __init__(self, pert_dict):
        """
        Constructor method

        Parameters
        ----------
        pert_dict : dict
            Dictionary containing the inputs and outputs from the transport calculation.

        """
        super().__init__(pert_dict)

        if self.calc_mode.split('-')[0] != 'trans':
            raise ValueError('Calculation mode for a Trans object should begin with "trans-"')

        self.temper = UnitsDict(units=self._pert_dict['trans'].pop('temperature units'))
        self.chem_pot = UnitsDict(units=self._pert_dict['trans'].pop('chemical potential units'))
        self.conc = UnitsDict(units=self._pert_dict['trans'].pop('concentration units'))
        self.cond = UnitsDict(units=self._pert_dict['trans'].pop('conductivity units'))
        self.mob = UnitsDict(units=self._pert_dict['trans'].pop('mobility units'))

        if 'Seebeck coefficient units' in self._pert_dict['trans'].keys():
            self.seebeck = UnitsDict(units=self._pert_dict['trans'].pop('Seebeck coefficient units'))
        else:
            self.seebeck = None

        if 'thermal conductivity units' in self._pert_dict['trans'].keys():
            self.thermal_cond = UnitsDict(units=self._pert_dict['trans'].pop('thermal conductivity units'))
        else:
            self.thermal_cond = None

        if 'magnetic field units' in self._pert_dict['trans'].keys():
            self.bfield = UnitsDict(units=self._pert_dict['trans'].pop('magnetic field units'))
        else:
            self.bfield = None

        num_config = self._pert_dict['trans'].pop('number of configurations')
        trans_dat = self._pert_dict['trans'].pop('configuration index')

        if 'number of iterations' in trans_dat[1].keys():
            self.cond_iter = UnitsDict(units=self.cond.units)
        else:
            self.cond_iter = None

        for config_idx in trans_dat.keys():
            self.temper[config_idx] = trans_dat[config_idx].pop('temperature')
            self.chem_pot[config_idx] = trans_dat[config_idx].pop('chemical potential')
            self.conc[config_idx] = trans_dat[config_idx].pop('concentration')
            self.cond[config_idx] = np.array(trans_dat[config_idx].pop('conductivity')['tensor'])
            self.mob[config_idx] = np.array(trans_dat[config_idx].pop('mobility')['tensor'])

            if self.seebeck is not None:
                self.seebeck[config_idx] = np.array(trans_dat[config_idx].pop('Seebeck coefficient')['tensor'])
            
            if self.thermal_cond is not None:
                self.seebeck[config_idx] = np.array(trans_dat[config_idx].pop('thermal conductivity')['tensor'])

            if self.bfield is not None:
                self.bfield[config_idx] = np.array(trans_dat[config_idx].pop('magnetic field'))

            if self.cond_iter is not None:
                num_iter = trans_dat[config_idx].pop('number of iterations')
                iteration_dat = trans_dat[config_idx].pop('iteration')
                conductivity_iter = {}

                for iteration in iteration_dat.keys():
                    conductivity_iter[iteration] = iteration_dat[iteration]['conductivity']['tensor']

                self.cond_iter[config_idx] = conductivity_iter



"""
This module contains functions to compare hdf5 files

"""
import h5py
import numpy as np
from perturbopy.test_utils.run_test.run_utils import get_tol


def equal_scalar(scalar1, scalar2, key, ig_n_tol):
    """
    Determines if two scalars contain the same value

    Parameters
    ----------
    scalar1 : numpy.dtype
       first  scalar
    scalar2 : numpy.dtype
       second scalar
    key : str
       key associated with this scalar
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values

    Returns
    -------
    equal_value : bool
       boolean specifying if both scalars contain the same values

    """
    # check that scalar1 and scalar2 are Numbers
    errmsg = ('scalar1/2 are not float64 or int32')
    assert (
           (scalar1.dtype == 'int32' and scalar2.dtype == 'int32')
           or (scalar1.dtype == 'float64' and scalar2.dtype == 'float64')
           ), errmsg

    atol, rtol = get_tol(ig_n_tol, key)

    equal_value = np.allclose(np.array(scalar1),
                              np.array(scalar2),
                              atol=atol,
                              rtol=rtol,
                              equal_nan=True)

    diff = np.abs(scalar1 - scalar2)

    if np.abs(scalar1) > 1e-10:
        rdiff = np.abs((scalar2 - scalar1) / scalar1)
        diff_str = f'{diff:.1e}, {rdiff*100:.1e}%, {scalar1 = }, {scalar2 = }'

    else:
        diff_str = f'{diff:.1e}'

    return equal_value, diff_str


def equal_ndarray(ndarray1, ndarray2, key, ig_n_tol):
    """
    Determines if two ndarrays contain the same values

    Parameters
    ----------
    ndarray1 : numpy.ndarray
       first numpy ndarray
    ndarray2 : numpy.ndarray
       second numpy ndarray
    key: str
       A key for the tolerance. If this key is not specified in the tolerance dict,
       a default tolerance will be applied
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values

    Returns
    -------
    equal_vlaues : bool
       boolean specifying if both ndarrays are equivalent

    """
    errmsg = ('ndarray1/2 are not ndarrays')
    assert isinstance(ndarray1, np.ndarray) and isinstance(ndarray2, np.ndarray), \
           errmsg

    errmsg = ('ndarray1/2 have different shapes')
    assert ndarray1.shape == ndarray2.shape, errmsg

    atol, rtol = get_tol(ig_n_tol, key)

    equal_value = np.allclose(ndarray1,
                              ndarray2,
                              atol=atol,
                              rtol=rtol,
                              equal_nan=True)

    diff = np.max(np.abs(ndarray1 - ndarray2))

    # find max percentage diff only if comparison fails
    if not equal_value:

        tmp_array = ndarray1
        tmp_array[np.abs(tmp_array) < 1e-10] = 1e-10

        idxmax_flat = np.argmax(np.abs((ndarray2 - ndarray1) / tmp_array))
        idxmax = np.unravel_index(idxmax_flat, ndarray1.shape)

        v1 = ndarray1[idxmax]
        v2 = ndarray2[idxmax]

        rdiff = np.abs((v2 - v1) / v1)
        diff_str = (f'{diff:.1e}, {rdiff*100:.1e}%, v1={v1:.1e}, v2={v2:.1e},'
                    f'atol={atol:.1e}, rtol={rtol:.1e},')

    else:
        diff_str = f'{diff:.1e}'

    return equal_value, diff_str


def equal_dict(dict1, dict2, ig_n_tol, path):
    r"""
    Determines if two dicts contain the same value
    for the same key

    .. note::

       Dict structure is assumed to be composed of
       other dictionaries, numpy.ndarray, numpy.int32 and
       numpy.float64

    Parameters
    ----------
    dict1 : dict
       first  dictionary
    dict2 : dict
       second dictionary
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values
    path : str
       pseudo path to this item being compared

    Returns
    -------
    equal_values : bool
       boolean specifying if both dicts contain the same keys and values

    """

    # check that dict1 and dict2 are dictionaries
    errmsg = ('dic1/2 are not dictionaries')
    assert isinstance(dict1, dict) and isinstance(dict2, dict), errmsg

    # check that dictionaries have the same keys
    errmsg = (f'dict1/2 found at {path} do not have the same keys')
    assert dict1.keys() == dict2.keys(), errmsg

    # total set of keys
    keys = set(dict1.keys())

    for key in dict1.keys():
        # remove 'ignore keys'
        if 'ignore keywords' in ig_n_tol:
            if key in ig_n_tol['ignore keywords']:
                keys.remove(key)

    # a list of bool values
    equal_per_key = []

    for key in keys:
        errmsg = (f'dict1/2 values associated with key:{key} '
                  f'are not of the same type')
        assert type(dict1[key]) == type(dict2[key]), errmsg

        # pseudo path to current item being compared
        key_path = (f'{path}.{key}')
        if isinstance(dict1[key], dict):
            equal_value, diff = equal_dict(dict1[key], dict2[key], ig_n_tol, key_path)

        elif isinstance(dict1[key], np.ndarray):
            equal_value, diff = equal_ndarray(dict1[key], dict2[key], key, ig_n_tol)

        elif (dict1[key].dtype == 'int32' or dict1[key].dtype == 'float64'):
            equal_value, diff = equal_scalar(dict1[key], dict2[key], key, ig_n_tol)

        else:
            errmsg = (f'dict must only contain values of type dict, np.ndarray, np.int32,or np.float64 '
                      f'but found type {type(dict1[key])}')
            known_types_present = False
            assert known_types_present, errmsg

        equal_per_key.append(equal_value)

        if not equal_value:
            print(f'\n !!! discrepancy found at {key_path}')
            print(f' difference: {diff}')

    # equal dicts produce list of only bool=True
    nitems = len(equal_per_key)
    ncompared = sum(equal_per_key)

    equal_values  = (nitems == ncompared)

    if equal_values:
        diff = None
    else:
        diff = f'among {nitems} elements, {nitems - ncompared} failed comparison'

    return equal_values, diff


def hdf5_to_dict(file_path):
    r"""
    Read the hdf5 file and return it in the form of dictionary

    Parameters
    ----------
    file_path : str
        path to the hdf5 file

    Returns
    -------
    hdf5_dict : dict
       data from the hdf5 file in the form of dictionary

    """
    hdf5_dict = {}
    with h5py.File(file_path, 'r') as file:
        # Recursively traverse all groups and datasets within the file
        def traverse_datasets(name, node):
            if isinstance(node, h5py.Dataset):
                hdf5_dict[name] = node[()]
            elif isinstance(node, h5py.Group):
                for key in node.keys():
                    traverse_datasets(f"{name}/{key}", node[key])
        
        # Start traversing the file, starting with the root group
        for key in file.keys():
            traverse_datasets(key, file[key])
    
    return hdf5_dict


def equal_values(file1, file2, ig_n_tol):
    """
    Checks if two h5 files contain the same hierarchy/groups/datasets


    Parameters
    ----------
    file1 : str
       first  HDF5 file name
    file2 : str
       second  HDF5 file name
    ig_n_tol : dict
       dictionary of keywords and tolerances needed to make comparison on files

    Returns
    -------
    equal_values : bool
       boolean specifying if both h5 files contain the same information

    """

    h51_dict = hdf5_to_dict(file1)
    h52_dict = hdf5_to_dict(file2)

    if 'test keywords' in ig_n_tol:
        h51_del_keys = [key for key in h51_dict.keys() if key not in ig_n_tol['test keywords']]
        h52_del_keys = [key for key in h52_dict.keys() if key not in ig_n_tol['test keywords']]

        for key in h51_del_keys:
            del h51_dict[key]
        for key in h52_del_keys:
            del h52_dict[key]

        errmsg = ('no entries left in dict after applying \'test keywords\'')
        assert len(h51_dict) > 0, errmsg
        assert len(h52_dict) > 0, errmsg

    return equal_dict(h51_dict, h52_dict, ig_n_tol, 'top of h5')



import numpy as np


class DynaIndivRun():
    """
    Class representation of a dynamics run

    Attributes
    ----------
    num_steps : int
        Number of time steps

    time_step : float
        Time step in fs

    snap_t : np.ndarray
        Distribution function computed at each time step

    efield : np.ndarray
        Electric field assumed during the calculation. Default value is [0, 0, 0].
    """

    def __init__(self, num_steps, time_step, snap_t, time_units='fs', efield=None):
        """
        Constructor method
        
        Parameters
        ----------
        num_steps : int
        time_step : float
            Time step in fs
        snap_t : array_like

        """

        self.num_steps = num_steps
        self.time_step = time_step
        self.snap_t = snap_t
        self.efield = efield



"""
   Set up the environment for Perturbo runs.
"""
import os
import shutil


def run_from_config_machine(config_machine, step):
    """
    Check if the executional command defined for the step of computation
    
    Parameters
    ----------
    config_machine : dict
        dictionary with executional commands for the test steps
    step : str
        name of running step ('scf', 'nscf', etc.)
    Returns

    Returns
    -------
    run : str
       string containing a command to run this step

    """

    # Read the perturbo_run variable from the environment
    try:
        run = config_machine['comp_info'][step]['exec']
    except KeyError:
        errmsg = (f'To run {step} as a part of the testsuite,\n'
                  'the exec element must be set in the configurational file.\n'
                  'Example:\n'
                  'exec: srun -n 64 pw.x -npools 8"'
                 )
        raise AttributeError(errmsg)

    return run
    

def copy_folder_with_softlinks(src, dst, perturbo_scratch_dir_prefix=None, test_name=None, second_run=False):
    """
    Copy the src directory to the dst directory, except the epr file - for it softlink is created.

    Parameters
    ----------
    src : str
        folder with computational info, which is copied
    
    dst : str
        folder, where the files will be copied

    perturbo_scratch_dir_prefix : str or None
        path to the scratch folder with computational files and results, None by default

    test_name : str or None
        name of test, None by default
    second_run : bool
        check that this is second run of perturbo (for qe2pert testing)

    Returns
    -------
    None
    """
    os.makedirs(dst)

    # Get the list of files in the source folder
    for root, dirs, files in os.walk(src):
        
        # Determine the relative path of the current directory
        relative_path = os.path.relpath(root, src)
        
        # Create the corresponding subdirectory in the destination folder
        dst_subfolder = os.path.join(dst, relative_path)
        if not os.path.exists(dst_subfolder):
            os.makedirs(dst_subfolder)

        for file_name in files:
            src_file_path = os.path.join(root, file_name)
            dst_file_path = os.path.join(dst_subfolder, file_name)

            if file_name.endswith("epr.h5"):
                # Create a softlink for files ending with "epr.h5"
                if not second_run:
                    # if it's first run of perturbo - simply make softlink of file
                    os.symlink(src_file_path, dst_file_path)
                else:
                    # if it's second run - softlink the file from our previous calculations
                    softlink_epr_files(perturbo_scratch_dir_prefix, test_name, dst_file_path, file_name)
            else:
                # Copy other files
                shutil.copy2(src_file_path, dst_file_path)
                

def perturbo_scratch_dir_config(source_folder, perturbo_inputs_dir_path, test_name, config_machine, test_case='perturbo', rm_preexist_dir=True):
    """
    Check if the PERT_SCRATCH variable is written in the config_machine file.
    If not - use default location "/PERT_SCRATCH".
    After it, this programm make the copy from folder with tests from either epr_computation or inputs folder.
    

    Parameters
    ----------
    source_folder : str
        path of source directory

    perturbo_inputs_dir_path : str
        folder with all input files for the test

    test_name : str
        name of test

    config_machine : dict
        dictionary with computational information, which we'll use in this set of computations.

    test_case : str
        define what type of the test we run - for perturbo testing or for the
        qe2pert testing.

    rm_preexist_dir : bool
        whether to remove dir if it preexists

    Returns
    -------
    perturbo_scratch_dir : str
        string containing the path to generate dir named tmp_test_name in which
        outputs for test_name will be generated

    """
    # Read the perturbo_run variable from the environment
    perturbo_scratch_dir_prefix   = os.path.join(source_folder, "PERT_SCRATCH")
    try:
        perturbo_scratch_dir_prefix = os.path.join(source_folder, config_machine['PERT_SCRATCH'])
    except KeyError:
        print(f'PSCRATCH not set in the config_machine. using default location -  {perturbo_scratch_dir_prefix}')

    # if folder already exist and we don't want to change it - simply pass rest of the function
    perturbo_scratch_dir = os.path.join(perturbo_scratch_dir_prefix, test_case, test_name)
    if not rm_preexist_dir:
        return perturbo_scratch_dir

    # else - copy over input files to scratch dir
    src = perturbo_inputs_dir_path
    dst = perturbo_scratch_dir
    if os.path.isdir(dst):
        print(f'\n directory {dst} exists. Removing this directory ...\n')
        shutil.rmtree(dst)
        if test_case == 'perturbo' or test_case == 'epr_calculation':
            copy_folder_with_softlinks(src, dst)
        elif test_case == 'perturbo_for_qe2pert':
            copy_folder_with_softlinks(src, dst, perturbo_scratch_dir_prefix, test_name, second_run=True)
    else:
        if test_case == 'perturbo' or test_case == 'epr_calculation':
            copy_folder_with_softlinks(src, dst)
        elif test_case == 'perturbo_for_qe2pert':
            copy_folder_with_softlinks(src, dst, perturbo_scratch_dir_prefix, test_name, second_run=True)

    return perturbo_scratch_dir
        
   
def softlink_epr_files(perturbo_scratch_dir_prefix, test_name, dst, file_name):
    """
    Make a softlink from scratch folder to the computed epr file

    Parameters
    ----------
    perturbo_scratch_dir_prefix : str
        path to the scratch folder with computational files and results

    test_name : str
        name of test

    dst : str
        folder where the softlink will be placed
    
    file_name : str
        name of the linked epr-file

    Raises
    ------
    ValueError
       if corresponding epr-file wasn't found or calculated.

    Returns
    -------
    None

    """
    epr_name = test_name[:test_name.find('-')]
    epr_address = os.path.join(perturbo_scratch_dir_prefix, 'epr_calculation', epr_name, 'qe2pert')
    # copy from our previous computation to the current computation folder
    try:
        src = os.path.abspath(os.path.join(epr_address, file_name))
        os.symlink(src, dst)
    except FileNotFoundError:
        raise FileNotFoundError(f"Ephr-file for {test_name} wasn't found or calculated")



import numpy as np
from perturbopy.postproc.utils.constants import conversion_factor, standardize_units_name


class UnitsDict(dict):
    """
    This is a class representation of a set of physical quantities with units organized in a dictionary.
    The physical quantities may either be arrays or floats.

    Attributes
    ----------
    data : dict
       Dictionary of floats or array_like types of physical quantities
    units : str {}
       The units of the physical quantities

    """
    def __init__(self, units, *args, **kwargs):
        """
        Constructor method

        """
        super().__init__(*args, **kwargs)

        self.units = units

    @classmethod
    def from_dict(cls, input_dict, units):
        """
        Class method to create a UnitsDict object from a Python dictionary and a set of units.

        """

        def convert_lists_to_numpy(data):
            """
            Helper function to recursively convert any lists in the dictionary to arrays

            """
            if isinstance(data, list):
                return np.array(data)
            if isinstance(data, dict):
                return {key: convert_lists_to_numpy(value) for key, value in data.items()}
            return data

        input_dict = convert_lists_to_numpy(input_dict)
        units_dict = cls(units)
        units_dict.update(input_dict)
        
        return units_dict



"""
   Run an executable for the testsuite.
"""
import numpy as np
import os
import sys
import shlex
import shutil
import subprocess
from perturbopy.io_utils.io import open_yaml
from perturbopy.test_utils.run_test.env_utils import run_from_config_machine
from perturbopy.test_utils.run_test.env_utils import perturbo_scratch_dir_config
from perturbopy.test_utils.run_test.run_utils import print_test_info, setup_default_tol
from perturbopy.test_utils.run_test.run_utils import ph_collection, define_nq_num


def run_perturbo(source_folder, perturbo_driver_dir_path, config_machine,
                 input_name='pert.in', output_name='pert.out'):
    """
    Function to run Perturbo and produce output files


    Parameters
    ----------
    source_folder : str
        path of source directory
    perturbo_driver_dir_path : str
        path to dir with pert.in file
    config_machine : dict
        dictionary with computational information, which we'll use in this set of computations.
    input_name : str, optional
        name of the input file, default: 'pert.in'
    output_name : str, optional
        name of the output file, default: 'pert.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'perturbo')

    run = f'{command} -i {input_name} | tee {output_name}'

    os.chdir(perturbo_driver_dir_path)

    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')

    print(f' === Running Perturbo === :\n {run}')
    sys.stdout.flush()

    subprocess.run(preliminary_commands(config_machine, 'perturbo') + run, shell=True)

    os.chdir(source_folder)
    

def preliminary_commands(config_machine, step):
    """
    Function which define all comands which you want to run before the qe2pert computations and
    each separate computation
    
    Parameters
    ----------
    config_machine : str
        dictionary, which include the list of running commands
    step : str
        name of the computational step which is computed ('scf', 'nscf', etc.)
    """

    print(' == Prel commands == :')
    list_of_coms = ''
    if 'prel_coms' in config_machine:
        for com in config_machine['prel_coms']:
            print(f' ======= Run ======= :\n {com}')
            sys.stdout.flush()
            list_of_coms += f'{com}\n'
    if 'prel_coms' in config_machine['comp_info'][step]:
        for com in config_machine['comp_info'][step]['prel_coms']:
            print(f' ======= Run ======= :\n {com}')
            sys.stdout.flush()
            list_of_coms += f'{com}\n'
    return list_of_coms
    

def run_scf(source_folder, work_path, config_machine, input_name='scf.in', output_name='scf.out'):
    """
    Function for scf calculation

    Parameters
    ----------
    source_folder : str
        path of source directory
    work_path : str
        path to dir with input file, where we'll run the calculations
    config_machine : dict
        dictionary, which include the commands for scf calculation
    input_name : str, optional
        name of the input file, default: 'scf.in'
    output_name : str, optional
        name of the output file, default: 'scf.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'scf')
    run = f'{command} -i {input_name} | tee {output_name}'

    os.chdir(f'{work_path}/pw-ph-wann/scf/')

    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')

    print(f' === Running scf === :\n {run}')
    sys.stdout.flush()

    subprocess.run(preliminary_commands(config_machine, 'scf') + run, shell=True)

    os.chdir(source_folder)
    

def run_phonon(source_folder, work_path, config_machine, prefix, input_name='ph.in', output_name='ph.out'):
    """
    Function for nscf calculation

    Parameters
    ----------
    source_folder : str
        path of source directory
    work_path : str
        path to dir with input file, where we'll run the calculations
    config_machine : dict
        dictionary, which include the commands for phonon calculation
    prefix : str
        prefix which we use for the filenames
    input_name : str, optional
        name of the input file, default: 'ph.in'
    output_name : str, optional
        name of the output file, default: 'ph.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'phonon')
    run = f'{command} -i {input_name} | tee {output_name}'

    os.chdir(f'{work_path}/pw-ph-wann/phonon/')
    softlink = '../scf/tmp'
    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')
    print(f'\n =Link tmp from scf= :\n {softlink}')
    sys.stdout.flush()
    os.symlink(softlink, 'tmp')

    print(f' == Running Phonon = :\n {run}')
    sys.stdout.flush()

    subprocess.run(preliminary_commands(config_machine, 'phonon') + run, shell=True)
    
    nq_num = define_nq_num(output_name)
    print(f' == Collect files == :\n ph_collection({prefix},{nq_num})')
    sys.stdout.flush()
    ph_collection(prefix, nq_num)
    
    os.chdir(source_folder)

    
def run_nscf(source_folder, work_path, config_machine, input_name='nscf.in', output_name='nscf.out'):
    """
    Function for nscf calculation

    Parameters
    ----------
    source_folder : str
        path of source directory
    work_path : str
        path to dir with input file, where we'll run the calculations
    config_machine : dict
        dictionary, which include the commands for nscf calculation
    input_name : str, optional
        name of the input file, default: 'nscf.in'
    output_name : str, optional
        name of the output file, default: 'nscf.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'nscf')
    run = f'{command} -i {input_name} | tee {output_name}'

    os.chdir(f'{work_path}/pw-ph-wann/nscf/')
    softlink = '../scf/tmp'
    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')
    print(f'\n =Link tmp from scf= :\n {softlink}')
    sys.stdout.flush()
    os.symlink(softlink, 'tmp')

    print(f' === Running nscf === :\n {run}')
    sys.stdout.flush()

    subprocess.run(preliminary_commands(config_machine, 'nscf') + run, shell=True)

    os.chdir(source_folder)
    

def run_wannier(source_folder, work_path, config_machine, prefix, input_name='pw2wan.in', output_name='pw2wan.out'):
    """
    Function for wannier90 calculation

    Parameters
    ----------
    source_folder : str
        path of source directory
    work_path : str
        path to dir with input file, where we'll run the calculations
    config_machine : dict
        dictionary, which include the commands for wannier calculation
    prefix : str
        prefix which we use for the filenames
    input_name : str, optional
        name of the input file, default: 'pw2wan.in'
    output_name : str, optional
        name of the output file, default: 'pw2wan.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'wannier90')
    run = f'{command} -pp {prefix}'
    
    # link the save-file from scf calculation
    os.chdir(f'{work_path}/pw-ph-wann/wann/')
    os.mkdir('tmp')
    softlink = f'../../scf/tmp/{prefix}.save'
    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')
    print(f'\n =Link tmp from scf= :\n {softlink}')
    sys.stdout.flush()
    os.symlink(softlink, f'tmp/{prefix}.save')

    # first run of wannier90
    print(f' === Running pp === :\n {run}')
    sys.stdout.flush()
    subprocess.run(preliminary_commands(config_machine, 'wannier90') + run, shell=True)
    
    # run of pw2wan
    command = run_from_config_machine(config_machine, 'pw2wannier90')
    run = f'{command} -i {input_name} | tee {output_name}'
    print(f' = Running pw2wan = :\n {run}')
    sys.stdout.flush()
    subprocess.run(preliminary_commands(config_machine, 'pw2wannier90') + run, shell=True)
    
    # second run of wannier90
    command = run_from_config_machine(config_machine, 'wannier90')
    run = f'{command} {prefix}'
    print(f' = Running Wannier= :\n {run}')
    sys.stdout.flush()
    subprocess.run(preliminary_commands(config_machine, 'wannier90') + run, shell=True)
    
    os.chdir(source_folder)
    

def run_qe2pert(source_folder, work_path, config_machine, prefix, input_name='qe2pert.in', output_name='qe2pert.out'):
    """
    Function for qe2pert calculation

    Parameters
    ----------
    source_folder : str
        path of source directory
    work_path : str
        path to dir with input file, where we'll run the calculations
    config_machine : dict
        dictionary, which include the commands for qe2pert calculation
    prefix : str
        prefix which we use for the filenames
    input_name : str, optional
       name of the input file, default: 'qe2pert.in'
    output_name : str, optional
       name of the output file, default: 'qe2pert.out'

    Returns
    -------
    None

    """

    command = run_from_config_machine(config_machine, 'qe2pert')
    run = f'{command} -i {input_name} | tee {output_name}'

    # link the save-file from scf calculation
    os.chdir(f'{work_path}/qe2pert/')
    os.mkdir('tmp')
    softlink = f'../../pw-ph-wann/nscf/tmp/{prefix}.save'
    print(f'\n ====== Path ======= :\n {os.getcwd()}\n')
    print(f'\n =Link tmp from nscf :\n {softlink}')
    sys.stdout.flush()
    os.symlink(softlink, f'tmp/{prefix}.save')
    
    # link rest files
    softlink = f'../pw-ph-wann/wann/{prefix}_u.mat'
    print(f'\n = Link rest files = :\n {softlink};')
    sys.stdout.flush()
    os.symlink(softlink, f'{prefix}_u.mat')
    softlink = f'../pw-ph-wann/wann/{prefix}_u_dis.mat'
    print(f'\n {softlink};')
    sys.stdout.flush()
    os.symlink(softlink, f'{prefix}_u_dis.mat')
    softlink = f'../pw-ph-wann/wann/{prefix}_centres.xyz'
    print(f'\n {softlink};')
    sys.stdout.flush()
    os.symlink(softlink, f'{prefix}_centres.xyz')

    # run qe2pert
    print(f' = Running qe2pert = :\n {run}')
    sys.stdout.flush()
    
    subprocess.run(preliminary_commands(config_machine, 'qe2pert') + run, shell=True)

    os.chdir(source_folder)


def get_test_materials(test_name, test_case, config_machine, source_folder):
    """
    Run one test:
       #. run perturbo.x to produce output files
       #. determine paths to files for comparison
       #. associated settings for file comparison with file paths

    Parameters
    ----------
    test_name : str
        name of test
    test_case : str
        define what type of the test we run - for perturbo testing or for the
        qe2pert testing.
    config_machine : str
        name of file with computational information, which we'll use in this set of computations.
        Should be in folder {source_folder}/config_machine
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -----
    ref_outs : list
       list of paths to reference files
    new_outs : list
       list of paths to outputted files
    igns_n_tols : list
       list of dictionaries containing the ignore keywords and tolerances needed to performance comparison of ref_outs and new_outs

    """
    # suffixes of paths needed to find driver/utils/references
    inputs_path_suffix = 'inputs/' + test_name
    ref_data_path_suffix = 'refs/' + test_name

    config_machine = open_yaml(os.path.join(source_folder, f'config_machine/{config_machine}'))

    # determine needed paths
    perturbo_inputs_dir_path = [x[0] for x in os.walk(source_folder) if x[0].endswith(inputs_path_suffix)][0]
    work_path                = perturbo_scratch_dir_config(source_folder, perturbo_inputs_dir_path, test_name, config_machine, test_case)
    ref_path                 = [x[0] for x in os.walk(source_folder) if x[0].endswith(ref_data_path_suffix)][0]

    # input yaml for perturbo job
    pert_input = open_yaml(f'{work_path}/pert_input.yml')
    # dictionary containing information about files to check
    test_files = pert_input['test info']['test files']
    # names of files to check
    out_files  = test_files.keys()
    # list of full paths to reference outputs
    ref_outs    = [ref_path + '/' + out_file for out_file in out_files]
    # list of full paths to new outputs
    new_outs    = [work_path + '/' + out_file for out_file in out_files]

    # print the test information before the run
    print_test_info(test_name, pert_input, test_type='perturbo')

    # run Perturbo to produce outputs
    run_perturbo(source_folder, work_path, config_machine)

    # list of dict. Each dict contains ignore keywords and
    # tolerances (information about how to compare outputs)
    igns_n_tols = [test_files[out_file] for out_file in out_files]

    igns_n_tols = setup_default_tol(igns_n_tols, test_case)

    return (ref_outs,
            new_outs,
            igns_n_tols)


def run_epr_calculation(epr_name, config_machine, source_folder):
    """
    Run one test:
        #. Run scf calculation
        #. Run phonon calculation
        #. Run nscf calculation
        #. Run wannier90 calculation
        #. Run qe2pert calculation

    Parameters
    ----------
    epr_name : str
        name of computed epr_name file
    config_machine : str
        name of file with computational information, which we'll use in this set of computations.
        Should be in folder {source_folder}/config_machine.
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -----
    None
    """
    # suffixes of paths needed to find driver/utils/references
    inputs_path_suffix = f'epr_computation/{epr_name}'
    config_machine = open_yaml(os.path.join(source_folder, f'config_machine/{config_machine}'))

    # determine needed paths
    inputs_dir_path = os.path.join(source_folder, inputs_path_suffix)
    work_path = perturbo_scratch_dir_config(source_folder, inputs_dir_path, epr_name, config_machine, test_case='epr_calculation')
    
    # open input yaml-files with supplementary info
    # and computational commands
    input_yaml = open_yaml(os.path.join(source_folder, 'epr_info.yml'))

    # print the test information before the run
    print_test_info(epr_name, input_yaml, test_type='qe2pert')
    
    # define the prefix - we'll need to have it in the later computations
    prefix = input_yaml[epr_name]['prefix']

    # run scf
    run_scf(source_folder, work_path, config_machine)
    
    # run phonon
    run_phonon(source_folder, work_path, config_machine, prefix)

    # run nscf
    run_nscf(source_folder, work_path, config_machine)
    
    # run wannier90
    run_wannier(source_folder, work_path, config_machine, prefix)
    
    # run qe2pert
    run_qe2pert(source_folder, work_path, config_machine, prefix)

    return


def clean_test_materials(test_name, new_outs, config_machine, source_folder):
    """
    clean one test:
       #. removes new files and dirs produced by test

    Parameters
    ----------
    test_name : str
       name of test
    new_outs : list
       list of paths to produced outputs
    config_machine : dict
        dictionary with computational information, which we'll use in this set of computations.
    source_folder : str
        path of source directory

    Returns
    -----
    None

    """
    # suffixes of paths needed to find driver/utils/references
    inputs_path_suffix = 'inputs/' + test_name

    config_machine = open_yaml(os.path.join(source_folder, f'config_machine/{config_machine}'))

    # determine paths
    perturbo_inputs_dir_path = [x[0] for x in os.walk(source_folder) if x[0].endswith(inputs_path_suffix)][0]
    work_path                = perturbo_scratch_dir_config(source_folder, perturbo_inputs_dir_path, test_name, config_machine, rm_preexist_dir=False)

    if os.path.isdir(work_path):
        print(f'\n === Test {test_name} passed ===\n\n Removing {work_path} ...')
        shutil.rmtree(work_path)

    return None
    

def clean_epr_folders(epr_failed, config_machine, keep_epr, keep_preliminary, source_folder):
    """
    Delete all temporary epr folders for the tests which were passed

    Parameters
    ----------
    epr_failed : list
        names of epr calculations, for which we obtained errors. The
        corresponding folders will be saved
    
    config_machine : dict
        dictionary with computational information, which we'll use in this set of computations.
    
    keep_epr : bool
        save all epr-files from the qe2pert testing
    
    keep_preliminary : bool
        save all preliminary files for epr calculation
    
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -----
    None

    """
    config_machine = open_yaml(os.path.join(source_folder, f'config_machine/{config_machine}'))

    # looking for the  location with the temporary files
    work_path   = os.path.join(source_folder, "PERT_SCRATCH")
    try:
        work_path    = os.path.join(source_folder, config_machine['PERT_SCRATCH'])
    except KeyError:
        print(f'PERT_SCRATCH not set in the config_machine. using default location -  {work_path}')
    epr_dict_path = os.path.join(source_folder, 'epr_info.yml')
    
    # set of all epr-files
    epr_full_list = [epr for epr in open_yaml(epr_dict_path)]
    
    # set of epr-files, for which tests have passed succescfully - we can delete them
    deleting_epr = list(set(epr_full_list) - set(epr_failed))
    print('\n == Tests finished ==\n\n')
    if not keep_preliminary:
        # if keep_preliminary, simply pass this function - we'll save all
        # files in this case

        if keep_epr:
            # if keep only epr, collect them in a separate new folder:
            dst = os.path.join(work_path, 'collected_epr')
            os.mkdir(dst)
            
            # collect all epr-files
            for epr in epr_full_list:
                src = os.path.join(work_path, 'epr_calculation', epr, 'qe2pert')
                
                # only if the directory with the epr-file exist:
                if os.path.isdir(src):
                    file_list = os.listdir(src)
                    
                    # looking for the epr-files in the files list
                    for file_name in file_list:
                        if file_name.endswith('epr.h5'):
                            full_src = os.path.join(src, file_name)
                            full_dst = os.path.join(dst, file_name)

                            # copy this file in the new folder
                            shutil.copy2(full_src, dst)

        # last steps in both cases - delete all computational folders
        for epr in deleting_epr:
            del_dir = os.path.join(work_path, 'epr_calculation', epr)
            if os.path.isdir(del_dir):
                print(f'Removing {del_dir} ... \n')
                shutil.rmtree(del_dir)



import numpy as np
import os
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.io_utils.io import open_yaml, open_hdf5, close_hdf5


class DynaPP(CalcMode):
    """
    Class representation of a Perturbo dynamics-pp calculation.
    
    Attributes
    ----------

    times : np.ndarray
        Array of all the times from the dynamics-run simulations.
    
    time_units : str
        units for the times attribute

    energy_grid : np.ndarray
        Array of energies in the energy window for the calculation
    
    energy_units : str
        Units for the energy_grid attribute

    popu : np.ndarray
        Array of shape num_times x num_energies containing carrier populations over the energy_grid and times.

    """

    def __init__(self, popu_file, pert_dict):
        """
        Constructor method
        
        Parameters
        ----------
        pert_dict : dict
            Dictionary containing the inputs and outputs from the dynamics-pp calculation.
        """
        super().__init__(pert_dict)

        if self.calc_mode != 'dynamics-pp':
            raise ValueError('Calculation mode for a DynamicsPPCalcMode object should be "dynamics-pp"')

        self.time_units = 'fs'
        self.times = popu_file['times_fs'][()]
        self.energy_units = 'ev'
        self.energy_grid = popu_file['energy_grid_ev'][()]

        self.popu = np.zeros((len(self.energy_grid), len(self.times)))

        for itime, time in enumerate(self.times):
            self.popu[:, itime] = popu_file['energy_distribution'][f'popu_t{itime}'][()]

    @classmethod
    def from_hdf5_yaml(cls, popu_path, yaml_path='pert_output.yml'):
        """
        Class method to create a DynamicsRunCalcMode object from the HDF5 file and YAML file
        generated by a Perturbo calculation

        Parameters
        ----------
        popu_path : str
           Path to the HDF5 file generated by a dynamics-pp calculation
        yaml_path : str, optional
           Path to the YAML file generated by a dynamics-pp calculation

        Returns
        -------
        dyanamics_pp : DynaPP
           The DynaPP object generated from the HDF5 and YAML files

        """

        if not os.path.isfile(popu_path):
            raise FileNotFoundError(f'File {popu_path} not found')
        if not os.path.isfile(yaml_path):
            raise FileNotFoundError(f'File {yaml_path} not found')

        popu_file = open_hdf5(popu_path)
        yaml_dict = open_yaml(yaml_path)

        return cls(popu_file, yaml_dict)



"""
This module contains functions to compare YAML files

"""
from numbers import Number
import numpy as np
import perturbopy.test_utils.compare_data.h5 as ch5
from perturbopy.test_utils.run_test.run_utils import get_tol
from perturbopy.io_utils.io import open_yaml


def equal_scalar(scalar1, scalar2, key, ig_n_tol):
    """
    Determines if two scalars contain the same value

    Parameters
    ----------
    scalar1 : numpy.dtype
       first  scalar
    scalar2 : numpy.dtype
       second scalar
    key : str
       key associated with this scalar
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values

    Returns
    -------
    equal_value : bool
       boolean specifying if both scalars contain the same values

    """
    # check that scalar1 and scalar2 are Numbers
    errmsg = ('scalar1/2 are not Numbers')
    assert isinstance(scalar1, Number) and isinstance(scalar2, Number), errmsg

    atol, rtol = get_tol(ig_n_tol, key)

    equal_value = np.allclose(np.array(scalar1),
                              np.array(scalar2),
                              atol=atol,
                              rtol=rtol,
                              equal_nan=True)

    diff = np.abs(scalar1 - scalar2)

    if np.abs(scalar1) > 1e-10:
        rdiff = np.abs((scalar2 - scalar1) / scalar1)
        diff_str = f'{diff:.1e}, {rdiff*100:.1e}%, {scalar1 = }, {scalar2 = }'

    else:
        diff_str = f'{diff:.1e}'

    return equal_value, diff_str


def equal_list(list1, list2, key, ig_n_tol, path):
    """
    Determines if two lists contain the same values

    Parameters
    ----------
    list1 : list
       first  list
    list2 : list
       second list
    key: str
       A key for the tolerance. If this key is not specified in the tolerance dict,
       a default tolerance will be applied
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values
    path : str
       pseudo path to this item being compared

    Returns
    -------
    equal_vlaues : bool
       boolean specifying if both lists are equivalent

    """
    # check that list1 and list2 are lists
    errmsg = ('list1/2 are not lists')
    assert isinstance(list1, list) and isinstance(list2, list), errmsg

    errmsg = ('list1/2 are not the same length')
    assert len(list1) == len(list2), errmsg
    indices = range(len(list1))

    # check if lists can be converted to ndarray
    if (all(isinstance(x, Number) for x in list1) and all(isinstance(x, Number) for x in list2)):

        # compare lists as nparrays for speed up
        return ch5.equal_ndarray(np.array(list1),
                                 np.array(list2),
                                 key,
                                 ig_n_tol)

    # a list of bool values
    equal_per_item = []

    for item1, item2, index in zip(list1, list2, indices):
        errmsg = ('list1/2 values'
                  'are not of the same type')
        assert type(item1) == type(item2), errmsg

        # pseudo path to current item being compared
        item_path = (f'{path}.list[{index}]')
        if isinstance(item1, dict):
            equal_value, diff = equal_dict(item1, item2, ig_n_tol, item_path)

        elif isinstance(item1, list):
            equal_value, diff = equal_list(item1, item2, key, ig_n_tol, item_path)

        elif isinstance(item1, Number):
            equal_value, diff = equal_scalar(item1, item2, key, ig_n_tol)

        elif isinstance(item1, str):
            equal_value = item1 == item2
            diff = f'{item1} {item2}'

        elif isinstance(item1, type(None)):
            equal_value = item1 == item2
            diff = None

        else:
            errmsg = ('list must only contain values of type dict, list, scalar, None, or str')
            known_types_present = False
            assert known_types_present, errmsg

        equal_per_item.append(equal_value)
        if not equal_value:
            print(f'\n !!! discrepancy found at {item_path}')
            print(f' difference: {diff}')

    # equal dicts produce list of only bool=True
    nitems = len(equal_per_item)
    ncompared = sum(equal_per_item)

    equal_values  = (nitems == ncompared)

    if equal_values:
        diff = None
    else:
        diff = f'among {nitems} elements, {nitems - ncompared} failed comparison'

    return equal_values, diff


def equal_dict(dict1, dict2, ig_n_tol, path):
    """
    Determines if two dicts contain the same value
    for the same key

    Parameters
    ----------
    dict1 : dict
       first  dictionary
    dict2 : dict
       second dictionary
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values
    path : str
       pseudo path to this item being compared

    Returns
    -------
    equal_vlaues : bool
       boolean specifying if both dicts contain the same keys and values

    """
    # check that dict1 and dict2 are dictionaries
    errmsg = (f'dict1/2 at {path} are not dictionaries')
    assert isinstance(dict1, dict) and isinstance(dict2, dict), errmsg

    # check that dictionaries have the same keys
    errmsg = (f'dict1/2 found at {path} do not have the same keys')
    assert dict1.keys() == dict2.keys(), errmsg

    # total set of keys
    keys = set(dict1.keys())

    for key in dict1.keys():
        # remove 'ignore keys'
        if 'ignore keywords' in ig_n_tol:
            if key in ig_n_tol['ignore keywords']:
                keys.remove(key)

    # a list of bool values
    equal_per_key = []

    for key in keys:
        errmsg = (f'dict1/2 values associated with key:{key} '
                  f'are not of the same type')
        assert type(dict1[key]) == type(dict2[key]), errmsg

        # pseudo path to current item being compared
        key_path = (f'{path}.{key}')
        if isinstance(dict1[key], dict):
            equal_value, diff = equal_dict(dict1[key], dict2[key], ig_n_tol, key_path)

        elif isinstance(dict1[key], list):
            equal_value, diff = equal_list(dict1[key], dict2[key], key, ig_n_tol, key_path)

        elif isinstance(dict1[key], Number):
            equal_value, diff = equal_scalar(dict1[key], dict2[key], key, ig_n_tol)

        elif isinstance(dict1[key], str):
            equal_value = dict1[key] == dict2[key]
            diff = f'{dict1[key]} {dict2[key]}'

        elif isinstance(dict1[key], type(None)):
            equal_value = dict1[key] == dict2[key]
            diff = None

        else:
            errmsg = (f'dict must only contain values of type dict, list, scalar, None, or str '
                      f'but found type {type(dict1[key])}')
            known_types_present = False
            assert known_types_present, errmsg

        equal_per_key.append(equal_value)
        if not equal_value:
            print(f'\n !!! discrepancy found at {key_path}')
            print(f' difference: {diff}')

    # equal dicts produce list of only bool=True
    nitems = len(equal_per_key)
    ncompared = sum(equal_per_key)

    equal_values  = (nitems == ncompared)

    if equal_values:
        diff = None
    else:
        diff = f'among {nitems} elements, {nitems - ncompared} failed comparison'

    return equal_values, diff


def equal_values(file1, file2, ig_n_tol):
    """
    Determines if two YAML files contain the same value
    for the same keys. keys must be in 'test keywords'

    Parameters
    ----------
    file1 : str
       first  YAML file name
    file2 : str
       second YAML file name
    ig_n_tol : dict
       dictionary of keywords and tolerances needed to make comparison on files

    Returns
    -------
    equal_vlaues : bool
       boolean specifying if both YAML files contain the same keys and values

    """
    yaml1_dict = open_yaml(file1)
    yaml2_dict = open_yaml(file2)

    if 'test keywords' in ig_n_tol:
        yaml1_del_keys = [key for key in yaml1_dict.keys() if key not in ig_n_tol['test keywords']]
        yaml2_del_keys = [key for key in yaml2_dict.keys() if key not in ig_n_tol['test keywords']]

        for key in yaml1_del_keys:
            del yaml1_dict[key]
        for key in yaml2_del_keys:
            del yaml2_dict[key]

        errmsg = ('no entries left in dict after applying \'test keywords\'')
        assert len(yaml1_dict) > 0, errmsg
        assert len(yaml2_dict) > 0, errmsg

    return equal_dict(yaml1_dict, yaml2_dict, ig_n_tol, 'top of yaml')



import numpy as np
import os
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.postproc.dbs.recip_pt_db import RecipPtDB
from perturbopy.postproc.calc_modes.dyna_indiv_run import DynaIndivRun
from perturbopy.io_utils.io import open_yaml, open_hdf5, close_hdf5
from perturbopy.postproc.utils.timing import Timing, TimingGroup
from perturbopy.postproc.dbs.units_dict import UnitsDict


class DynaRun(CalcMode):
    """
    Class representation of a Perturbo dynamics-run calculation.

    Attributes
    ----------
    kpt : RecipPtDB
       Database for the k-points used in the bands calculation.
    bands : UnitsDict
       Database for the band energies computed by the bands calculation.
    num_runs : int
        Number of separate simulations performed
    _dat : dict
        Python dictionary of DynaIndivRun objects containing results from each simulation
    """

    def __init__(self, cdyna_file, tet_file, pert_dict):
        """
        Constructor method

        Parameters
        ----------
        pert_dict : dict
            Dictionary containing the inputs and outputs from the dynamics-run calculation.

        """
        
        self.timings = TimingGroup("dynamics-run")

        super().__init__(pert_dict)

        if self.calc_mode != 'dynamics-run':
            raise ValueError('Calculation mode for a DynamicsRunCalcMode object should be "dynamics-run"')

        kpoint = np.array(tet_file['kpts_all_crys_coord'][()])

        self.kpt = RecipPtDB.from_lattice(kpoint, "crystal", self.lat, self.recip_lat)
        self.bands = UnitsDict.from_dict

        energies = np.array(cdyna_file['band_structure_ryd'][()])
        energies_dict = {i + 1: np.array(energies[:, i]) for i in range(0, energies.shape[1])}
        self.bands = UnitsDict.from_dict(energies_dict, 'Ry')

        self._data = {}

        self.num_runs = cdyna_file['num_runs'][()]

        with self.timings.add('iterate_dyna') as t:

            for irun in range(1, self.num_runs + 1):
                dyn_str = f'dynamics_run_{irun}'

                num_steps = cdyna_file[dyn_str]['num_steps'][()]
                time_step = cdyna_file[dyn_str]['time_step_fs'][()]

                # a dynamics run must have at least one snap
                numk, numb = cdyna_file[dyn_str]['snap_t_1'][()].shape

                snap_t = np.zeros((numb, numk, num_steps), dtype=np.float64)

                for itime in range(num_steps):
                    snap_t[:, :, itime] = cdyna_file[dyn_str][f'snap_t_{itime+1}'][()].T

                # Get E-field, which is only present if nonzero
                if "efield" in cdyna_file[dyn_str].keys():
                    efield = cdyna_file[dyn_str]["efield"][()]
                else:
                    efield = np.array([0.0, 0.0, 0.0])

                self._data[irun] = DynaIndivRun(num_steps, time_step, snap_t, time_units='fs', efield=efield)

    @classmethod
    def from_hdf5_yaml(cls, cdyna_path, tet_path, yaml_path='pert_output.yml'):
        """
        Class method to create a DynamicsRunCalcMode object from the HDF5 file and YAML file
        generated by a Perturbo calculation

        Parameters
        ----------
        cdyna_path : str
           Path to the HDF5 file generated by a dynamics-run calculation
        tet_path : str
           Path to the HDF5 file generated by the setup calculation required before the dynamics-run calculation
        yaml_path : str, optional
           Path to the YAML file generated by a dynamics-run calculation

        Returns
        -------
        dyanamics_run : DynamicsRunCalcMode
           The DynamicsRunCalcMode object generated from the HDF5 and YAML files

        """

        if not os.path.isfile(yaml_path):
            raise FileNotFoundError(f'File {yaml_path} not found')
        if not os.path.isfile(cdyna_path):
            raise FileNotFoundError(f'File {cdyna_path} not found')
        if not os.path.isfile(tet_path):
            raise FileNotFoundError(f'File {tet_path} not found')

        yaml_dict = open_yaml(yaml_path)
        cdyna_file = open_hdf5(cdyna_path)
        tet_file = open_hdf5(tet_path)

        return cls(cdyna_file, tet_file, yaml_dict)

    def __getitem__(self, index):
        """
        Method to index the DynamicsRunCalcMode object

        Parameters
        ----------
        index : int
            The dynamics run requested, indexing starting at 1

        Returns
        -------
        dynamics_run: DynamicsRun
           Object containing information for the dynamics run

        """
        if index <= 0 or index > len(self._data):
            raise IndexError("Index out of range")

        return self._data[index]

    def __len__(self):
        """
        Method to get the number of runs in DynamicsRunCalcMode object

        Returns
        -------
        num_runs : int
            Number of runs
        """

        return self.num_runs

    def get_info(self):
        """
        Method to get overall information on dynamics runs
        """

        print(f"\nThis simulation has {self.num_runs} runs")

        for irun, dynamics_run in self._data.items():

            print(f"{'Dynamics run':>30}: {irun}")
            print(f"{'Number of steps':>30}: {dynamics_run.num_steps}")
            print(f"{'Time step (fs)':>30}: {dynamics_run.time_step}")
            print(f"{'Electric field (V/cm)':>30}: {dynamics_run.efield}")
            print("")



import numpy as np
import pytest
import math
import perturbopy.postproc as ppy


@pytest.mark.parametrize("test_points, expected_points", [

                        ([3, 3, 2], np.array([[3], [3], [2]])),
                        (np.array([3, 3, 2]), np.array([[3], [3], [2]])),
                        ([[1, 1, 1], [2, 2, 2], [3, 3, 3]], np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])),
                        ([[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]], np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])),
                        ([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]], np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])),
                        ([[1, 1], [2, 2], [3, 3]], np.array([[1, 1], [2, 2], [3, 3]])),
                        ([[1, 2, 3], [1, 2, 3]], np.array([[1, 1], [2, 2], [3, 3]]))

])
def test_reshape_points(test_points, expected_points):
    """"
    Method to test lattice.reshape_points function

    Parameters
    ----------
    test_points : array_like
       The array of points to reshape
    expected_points : array
       The expected reshaped array of points

    """

    if np.shape(test_points) == (3, 3):
        with pytest.warns(UserWarning, match='Reciprocal points are 3x3 array; assume points are already column-oriented'):
            test_points = ppy.lattice.reshape_points(test_points)
    else:
        test_points = ppy.lattice.reshape_points(test_points)

    assert(np.all(test_points == expected_points))
    assert(isinstance(test_points, np.ndarray))

    if len(np.shape(test_points)) == 1:
        assert(np.shape(test_points) == (3, 1))
    elif np.shape(test_points)[0] == 3:
        assert(np.shape(test_points) == (3, np.shape(test_points)[1]))
    elif np.shape(test_points)[1] == 3:
        assert(np.shape(test_points) == (3, np.shape(test_points)[0]))


@pytest.mark.parametrize("test_points_cart, test_points_cryst, test_lat, test_recip_lat", [
                        ([[0, 0, 0.5, 0.25, 0.25, 0.375], [0, 0.5, 0.5, 0.75, 0.625, 0.75], [0, 0.5, 0.5, 0.5, 0.625, 0.375]],
                         [[0, 0, 0.5, 0.5, 0.25, 0.75], [0, 1, 0.5, 1, 1, 0.75], [0, 0, 0.5, 0, 0.25, 0.]],
                         np.array([[0.5, 0.5, 0.0], [0.0, 0.5, 0.5], [0.5, 0.0, 0.5]]).T,
                         np.array([[1.0, 1.0, -1.0], [-1.0, 1.0, 1.0], [1.0, -1.0, 1.0]]).T)

])
def test_cryst2cart(test_points_cart, test_points_cryst, test_lat, test_recip_lat):
    """"
    Method to test lattice.cryst2cart function

    Parameters
    ----------
    test_points_cart : array_like
       The points in cartesian coordinates
    test_points_cryst : array_like
       The points in crystal coordinates
    test_lat : array_like
       The lattice vectors assumed in the conversion
    test_recip_lat : array_like
       The reciprocal lattice vectors assumed in the conversion

    """
    assert(np.allclose(ppy.lattice.cryst2cart(test_points_cryst, test_lat, test_recip_lat, real_space=True, forward=True), test_points_cart))
    assert(np.allclose(ppy.lattice.cryst2cart(test_points_cart, test_lat, test_recip_lat, real_space=True, forward=False), test_points_cryst))


@pytest.mark.parametrize("test_points1, test_points2, expected_distances", [

                        ([3, 3, 1], [3, 3, 1], 0),
                        ([[3], [2], [1]], [3, 2, 1], 0),
                        ([[3, 2, 1]], [3, 2, 1], 0),
                        ([1, 1, 1], [0, 2, 0], math.sqrt(3)),
                        ([[1, 2, 3], [4, 5, 6]], [1, 2, 3], [0, math.sqrt(27)]),
                        ([[1, 4], [2, 5], [3, 6]], [1, 2, 3], [0, math.sqrt(27)]),
                        ([[1, 2, 3], [4, 5, 6]], [[2, 3, 4], [5, 6, 7]], [math.sqrt(3), math.sqrt(3)]),
                        ([[1, 4], [2, 5], [3, 6]], [[2, 3, 4], [5, 6, 7]], [math.sqrt(3), math.sqrt(3)])

])
def test_compute_distances(test_points1, test_points2, expected_distances):
    """"
    Method to test lattice.compute_distances function

    Parameters
    ----------
    test_points1, test_points2 : array_like
       Arrays of test points between which to compute distances
    expected_distances : list
       The expected distances between the two arrays of test points

    """
    assert(np.all(np.isclose(ppy.lattice.compute_distances(test_points1, test_points2), expected_distances)))
    assert(np.all(np.isclose(ppy.lattice.compute_distances(test_points2, test_points1), expected_distances)))


@pytest.mark.parametrize("test_point, test_points_array, max_dist, nearest, expected", [
                        ([1, 1, 1], [[1, 1, 1, 0.1], [1, 1, 1, 0.1], [1, 1, 1, 0.2]], None, True, [0, 1, 2]),
                        ([1, 1, 1], [[1, 1, 1, 0.1], [1, 1, 1, 0.1], [.96, 0.96, 0.96, 0.2]], None, True, []),
                        ([1, 1, 1], [[1, 1, 1, 0.1], [1, 1, 1, 0.1], [.96, 0.96, 0.96, 0.2]], 0.0401, True, [0, 1, 2]),
                        ([1, 1, 1], [[1, 1, 1, 0.1], [1, 1, 1, 0.1], [.9, 0.9, 0.9, 0.2]], None, True, []),
                        ([1, 1, 1], [[1, 1, 1, 0.1], [1, 1, 1, 0.1], [.9, 0.9, 0.9, 0.2]], 0.2, True, [0, 1, 2]),
                        ([[1], [1], [1]], [[1, 1, 1, 0], [1, 1, 1, 0], [.9, 0.9, 0.9, 0]], 0.2, True, [0, 1, 2]),
                        ([1, 1, 1], [[1, 1, 1, 0], [1, 1, 1, 0], [1, 0.995, 0.995, 0]], None, True, [0]),
                        ([1, 1, 1], [[1, 1, 1, 0], [1, 1, 1, 0], [1, 0.995, 0.995, 0]], None, False, [0, 1, 2])
])
def test_find_point(test_point, test_points_array, max_dist, nearest, expected):
    """"
    Method to test lattice.find_point function

    Parameters
    ----------
    test_point : array_like
       The point to search
    test_points_array : array_like
       The array of points to be searched for matches
    max_dist : float
       The maximum distances between points to be considered a match
    nearest : bool
       Whether or not to return just the matches that are closest in distance, or all matches within max_dist
    expected : list
       The expected list of indices corresponding to points in test_points_array matching test_point

    """
    if max_dist is None:
        assert(np.all(ppy.lattice.find_point(test_point, test_points_array, nearest=nearest) == expected))
    else:
        assert(np.all(ppy.lattice.find_point(test_point, test_points_array, max_dist=max_dist, nearest=nearest) == expected))


@pytest.mark.parametrize("test_point, test_points_array, test_path_array, max_dist, expected", [
                        ([1, 2, 0], [[1, 1, 1], [1, 5, 6], [1, 2, 0], [3, 3, 3]], [1, 2, 3, 4], None, 3),
                        ([1, 2, 0], [[1, 1, 1], [1, 5, 6], [1.02, 2.01, 0.04], [3, 3, 3]], [1, 2, 3, 4], None, 3),
                        ([1, 2, 0], [[1, 1, 1], [1, 2, 0], [1, 2, 0], [3, 3, 3]], [0.1, 0.2, 0.3, 0.4], None, [0.2, 0.3]),
                        ([1, 2, 0], [[1, 1, 1], [1, 5, 6], [1, 2.1, 0], [3, 3, 3]], [1, 2, 3, 4], None, None),
                        ([1, 2, 0], [[1, 1, 1], [1, 5, 6], [1, 2.1, 0], [3, 3, 3]], [1, 2, 3, 4], 0.12, 3),
                        ([1, 2, 0], [[1, 1, 1], [1, 2, 0], [1, 2.1, 0], [3, 3, 3]], [1, 2, 3, 4], 0.12, 2)
])
def test_convert_point2path(test_point, test_points_array, test_path_array, max_dist, expected):
    """"
    Method to test lattice.convert_point2path function

    Parameters
    ----------
    test_point : array_like
       The point to convert to a path coordinate
    test_points_array : array_like
       The array of points
    test_path_array : array_like
       The path coordinates corresponding to points in test_points_array
    max_dist : float
       The maximum distances between test_point and a point in test_points_array to be considered a match
    expected : int
       The expected path coordinate or list of coordinates corresponding to the test_point

    """
    if max_dist is None:
        assert(np.all(ppy.lattice.convert_point2path(test_point, test_points_array, test_path_array) == expected))
    else:
        assert(np.all(ppy.lattice.convert_point2path(test_point, test_points_array, test_path_array, max_dist=max_dist) == expected))


@pytest.mark.parametrize("test_path, test_points_array, test_path_array, atol, expected", [
                        (3, [[1, 1, 1], [1, 5, 6], [1, 2, 0], [3, 3, 3]], [1, 2, 3, 4], None, [1, 2, 0]),
])
def test_convert_path2point(test_path, test_points_array, test_path_array, atol, expected):
    """"
    Method to test lattice.path2point function

    Parameters
    ----------
    test_path : float
       The point to convert to a path coordinate
    test_points_array : array_like
       The array of points
    test_path_array : array_like
       The path coordinates corresponding to points in test_points_array
    atol : float
       The maximum distances between test_path and the coordinates in test_path_array to be considered a match
    expected : array_like
       The expected point coordinates corresponding to the test_path coordinate

    """
    print(ppy.lattice.convert_path2point(test_path, test_points_array, test_path_array))
    assert(np.all(np.isclose(ppy.lattice.convert_path2point(test_path, test_points_array, test_path_array), expected)))



import numpy as np
import warnings

points_fcc = {'L': [0.5, 0.5, 0.5], 'X': [0.5, 0.0, 0.5], 'W': [0.5, 0.25, 0.75], 'K': [0.375, 0.375, 0.75], 'G': [0, 0, 0]}


def reshape_points(point_array):
    """
    Method to reshape reciprocal points into a 2d array of shape (3, N)
    such that reciprocal points are column-oriented
    - a list of reciprocal points will be transformed to an array
    - an array of reciprocal points with shape (N, 3) will be reshaped
      to shape (3, N) if N != 3
    - Note that reciprocal points with shape (3, 3) will be assumed to
      have the correct shape

    Parameters
    ----------
    point_array: array_like
        An array or list of reciprocal points with shape (3,N), (N,3), or (3,)

    Returns
    -------
    point_array: array
        An array of reciprocal points with shape (3,N)

    Raises
    ------
    ValueError
        If the shape of `point_array` is not (3, N), (N, 3), or (3,)

    Warns
    -----
    UserWarning
        If `point_array` is a 3x3 array, a warning is issued that the points are
        assumed to be column-oriented already.

    """

    point_array = np.array(point_array)
    point_array_shape = np.shape(point_array)

    if 3 not in point_array_shape:
        raise ValueError('Reciprocal points should be inputted as a 3xN array, where N is the number of points.')

    if len(point_array_shape) == 1:
        return np.reshape(point_array, (3, 1))

    elif point_array_shape[0] == 3:
        if point_array_shape[1] == 3:
            warnings.warn('Reciprocal points are 3x3 array; assume points are already column-oriented', UserWarning)
        return point_array

    else:
        return np.transpose(point_array)


def cryst2cart(point_array, lat, recip_lat, forward=True, real_space=True):
    """
    Method to convert points in real space or reciprocal space
    between crystal and cartesian coordinates, given the crystal
    lattice. Real space points in cartesian units have units of alat.
    Reciprocal points in cartesian coordinates have units of 2pi/alat.

    Parameters
    ----------
    point_array : array
        3xN Array of vectors [v1, v2, v3 ...] to be converted

    lat : array
        3x3 array of lattice vectors [v1, v2, v3] in units of alat

    recip_lat : array
        3x3 array of reciprocal lattice vectors [v1, v2, v3 in units of
        2pi/a

    forward : bool, optional
        If true, vectors will be converted from crystal to cartesian
        coordinates. If false, vectors will be converted from
        cartesian to crystal coordinates

    real_space : bool, optional
        If true, vectors are assumed to be in real space
        (i.e. atomic positions). If false, vectors are assumed to
        be in reciprocal space (i.e. reciprocal points).

    Returns
    -------
    converted_point_array : array
        Array containing the converted vectors [v1, v2, v3, ...]

    """

    if forward:
        if real_space:
            conversion_mat = lat
        else:
            conversion_mat = recip_lat

    else:
        if real_space:
            conversion_mat = np.transpose(recip_lat)
        else:
            conversion_mat = np.transpose(lat)

    converted_point_array = np.tensordot(conversion_mat, point_array, axes=1)

    return converted_point_array


def compute_distances(point_array1, point_array2):
    """
    Method to compute the distances between points

    Parameters
    ----------
    point_array1, point_array2 : array_like
        The points between which distances will be computed. They should have the same
        shape. Both will be reshaped to assume column-oriented points if needed. For
        example, arrays of shape (N, 3) will be reshaped to (3, N) if N != 3.

    Returns
    -------
    distances : array
        An array of distances between points in point_array1 and point_array2, computed
        across the columns.

    """
    point_array1 = reshape_points(point_array1)
    point_array2 = reshape_points(point_array2)

    if np.shape(point_array1) != np.shape(point_array2) and np.shape(point_array1) != (3, 1) and np.shape(point_array2) != (3, 1):
        raise ValueError("Shape of arrays should be the same, or one array should only contain one k-point.")

    distances = np.linalg.norm(point_array1 - point_array2, axis=0)

    return distances


def find_point(point, point_array, max_dist=0.025, nearest=True):
    """
    Method to find the index or indices of a particular point in an array of points.

    Parameters
    ----------
    point : array_like
        The point to be located in an array of points

    point_array : array_like
        The set of points that will be searched through

    max_dist : float, optional
        The maximum distance between the point to locate and the points identified as matches

    narest : bool, optional
        If True, only the nearest point, or points in the case of duplicate points, are returned (even if
        other points are within the max_dist)


    Returns
    -------
    points_indices: list
        The index or indices of the matching points in point_array

    """
    point = reshape_points(point)
    point_array = reshape_points(point_array)

    distances = compute_distances(point_array, point)

    if nearest:
        min_distance = np.min(distances)
        point_indices = np.where(np.isclose(distances, min_distance))[0]

        if min_distance <= max_dist:
            return point_indices
        else:
            return []

    else:
        return np.arange(len(distances))[distances <= max_dist]


def convert_point2path(point, point_array, path_array, max_dist=0.025, nearest=True):
    """
    Method to find the path coordinate corresponding to a particular point

    Parameters
    ----------
    point: array_like
        The point to be converted to a path coordinate

    point_array: array_like
        A set of points

    path_array: array_like
        The path corresponding to the points

    max_dist : float, optional
        The maximum distance between the point to locate and the matching points

    nearest : bool, optional
        If True, only the nearest point, or points in the case of repeated points, are returned (even if
        other points are within the max_dist)

    Returns
    -------
    path_coord: array
        The path coordinate(s) of the corresponding reciprocal space point(s)

    """
    point = reshape_points(point)
    point_array = reshape_points(point_array)
    path_array = np.array(path_array)

    point_indices = find_point(point, point_array, max_dist, nearest)

    if len(point_indices) == 0:
        return np.array([])
    else:
        path_coord = path_array[point_indices]
        return path_coord


def convert_path2point(path_coord, point_array, path_array, atol=1e-5, rtol=1e-2, nearest=True):
    """
    Method to find the point corresponding to a path coordinate

    Parameters
    ----------
    path_coord : float
        The path coordinate to be converted to a point

    point_array: array_like
        A set of points

    path_array: array_like
        The path corresponding to the points

    atol : float, optional
        The absolute tolerance between the path coordinate to locate and the matching path coordinates

    rtol : float, optional
        The relative tolerance between the path coordinate to locate and the matching path coordinates

    nearest : bool, optional
        If True, only the nearest path coordinate, or coordinates in the case of repeats, are returned (even if
        other points are within the absolute and relative tolerance)

    Returns
    -------
    point_array: list
        The reciprocal space point(s) of the corresponding path coordinate

    """
    point_array = reshape_points(point_array)
    path_array = np.array(path_array)

    distances = np.abs(path_array - path_coord)
    min_distance = np.amin(distances)

    if min_distance <= atol + rtol * path_coord:
        path_indices = np.where(np.isclose(distances, min_distance))[0]
        return np.reshape(point_array[:, path_indices], (3,))
    else:
        return np.array([])



from perturbopy.test_utils.run_test.test_driver import get_test_materials, run_epr_calculation
from perturbopy.test_utils.run_test.test_driver import clean_test_materials
from perturbopy.test_utils.compare_data.compare import equal_values

import os
import sys
import pytest
import subprocess


@pytest.mark.order(before="test_qe2pert")
def test_perturbo(test_name, config_machine, keep_perturbo, source_folder, test_case='perturbo'):
    """
    Driver to run the tests for the perturbo.x executable.

    Parameters
    -----
    test_name : str
        name of the folder inside the tests/ folder
    test_case : str
        define what type of the test we run - for perturbo testing or for the
        qe2pert testing.
    config_machine : str
        name of file with computational information, which we'll use in this set of computations.
        Should be in folder source_folder/config_machine
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -----
    None
    """

    # run test, get files paths, get comparisons settings
    (ref_outs,
     new_outs,
     igns_n_tols) = get_test_materials(test_name, test_case, config_machine, source_folder)

    # compare all files
    for ref_file, new_file, ign_n_tol in zip(ref_outs, new_outs, igns_n_tols):

        # print(f'\ncomparing files {ref_file_short} and {new_file_short}')
        print(f'\n comparing files: \n {ref_file}  {new_file}')

        errmsg = (f'files {ref_file} and {new_file} do not match')
        assert equal_values(ref_file, new_file, ign_n_tol), errmsg

    # clean up test materials
    if not keep_perturbo:
        clean_test_materials(test_name, new_outs, config_machine, source_folder)
    print('')
    
    
def test_qe2pert(test_name, run_qe2pert, config_machine, source_folder):
    """
    Driver to run the the qe2pert set of computations.

    Parameters
    -----
    test_name : str
        name of the computing epr-file
    run_qe2pert : bool
        do we run qe2pert testing or not
    config_machine : str
        name of file with computational information, which we'll use in this set of computations.
        Should be in folder source_folder/config_machine
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -----
    None
    """
    if not run_qe2pert:
        pytest.skip("Skipping by default, pass the --run_qe2pert arg in the command line for this test")
    run_epr_calculation(test_name, config_machine, source_folder)
    assert True
    

@pytest.mark.order(after="test_qe2pert")
def test_perturbo_for_qe2pert(test_name, run_qe2pert, config_machine, keep_perturbo, source_folder):
    """
    Second driver to run the tests for the perturbo.x executable.
    We call it only in the case if we call test_qe2pert as well

    Parameters
    -----
    test_name : str
        name of the folder inside the tests/ folder
    run_qe2pert : str
        do we run qe2pert testing or not
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)
    Returns
    -----
    None
    """
    if not run_qe2pert:
        pytest.skip("Skipping by default, pass the --run_qe2pert arg in the command line for this test")
    test_perturbo(test_name, config_machine, keep_perturbo, source_folder, test_case='perturbo_for_qe2pert')
    



"""
Open/close binary and ASCII files (HDF5, YAML, text inputs/outputs)

"""
from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import h5py


def open_yaml(file_name):
    """
    Load YAML file as dictionary

    Parameters
    ----------
    file_name : str
       name of YAML file to be loaded

    Returns
    -------
    yaml_dict : dict
       YAML file loaded as dict

    """
    with open(file_name, 'r') as file:
        yaml_dict = load(file, Loader=Loader)

    return yaml_dict


def open_hdf5(filename, mode='r'):
    hdf5_file = h5py.File(filename, mode)
    return hdf5_file


def close_hdf5(hdf5_file):
    hdf5_file.close()



import pytest
import perturbopy.postproc as ppy

points_cryst = [[0, 0, 0.5, 0.25, 0.25, 0.375], [0, 0.5, 0.5, 0.75, 0.625, 0.75], [0, 0.5, 0.5, 0.5, 0.625, 0.375]]
points_cart = [[0, 0, 0.5, 0.5, 0.25, 0.75], [0, 1, 0.5, 1, 1, 0.75], [0, 0, 0.5, 0, 0.25, 0.]]


@pytest.mark.parametrize("path, energies, energy_units, c, ls, energy_window", [
                        ([1, 2, 3, 4, 5], {1: [0.1, 0.2, 0.25, 0.3, 0.2], 2: [1.1, 1.2, 1.25, 1.3, 1.2]}, 'eV', 'k', '--', None),
                        ([1, 2, 3, 4, 5], {1: [0.1, 0.2, 0.25, 0.3, 0.2], 2: [1.1, 1.2, 1.25, 1.3, 1.2]}, 'eV', ['b', 'g'], '--', None),
                        ([1, 2, 3, 4, 5], {1: [0.1, 0.2, 0.25, 0.3, 0.2], 2: [1.1, 1.2, 1.25, 1.3, 1.2]}, 'eV', ['b', 'k'], ['-', '--'], None),
                        ([1, 2, 3, 4, 5], {1: [0.1, 0.2, 0.25, 0.3, 0.2], 2: [1.1, 1.2, 1.25, 1.3, 1.2]}, 'eV', 'k', '--', [0.2, 1]),
                        ([1, 2, 3, 4, 5], {1: [0.1, 0.2, 0.25, 0.3, 0.2], 2: [1.1, 1.2, 1.25, 1.3, 1.2], 3: [2, 2.3, 2.4, 2, 2]},
                            'eV', ['b', 'g'], ['--', ':'], None)
])
def test_plot_dispersion(plt, path, energies, energy_units, c, ls, energy_window, with_plt):
    """"
    Method to test plot_tools.plot_dispersion function

    Method to plot the dispersion (phonon dispersion or band structure).

    Parameters
    ----------
    path : array_like
       The array or list of path coordinates to be plotted on the x-axis

    energies : dict
       Dictionary of arrays of energies to plot, with keys labelling the band number or phonon mode.

    energy_units : str

    c : str, list
       Matplotlib color for plotting. If a list, different colors will be iterated through as each band is plotted.

    ls : str, list
       Matplotlib linestyle for plotting. If a list, different linestyles will be iterated through as each band is plotted.

    energy_window : tuple, optional
       Tuple with the lower and upper bound of the chosen energy window

    """

    if not with_plt:
        pytest.skip("Test requires pytest-plt")

    fig, ax = plt.subplots()
    ppy.plot_tools.plot_dispersion(ax, path, energies, energy_units, c, ls, energy_window)



"""
This is a module for creating plots based on Perturbo calculation results

"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from perturbopy.postproc.utils import lattice

plotparams = {'figure.figsize': (16, 9),
                     'axes.grid': False,
                     'lines.linewidth': 2.5,
                     'axes.linewidth': 1.1,
                     'lines.markersize': 10,
                     'xtick.bottom': True,
                     'xtick.top': True,
                     'xtick.direction': 'in',
                     'xtick.minor.visible': True,
                     'ytick.left': True,
                     'ytick.right': True,
                     'ytick.direction': 'in',
                     'ytick.minor.visible': True,
                     'figure.autolayout': False,
                     'mathtext.fontset': 'dejavusans',
                     'mathtext.default': 'it',
                     'xtick.major.size': 4.5,
                     'ytick.major.size': 4.5,
                     'xtick.minor.size': 2.5,
                     'ytick.minor.size': 2.5,
                     'legend.handlelength': 3.0,
                     'legend.shadow': False,
                     'legend.markerscale': 1.0,
                     'font.size': 20}

points_fcc = {'L': [0.5, 0.5, 0.5], 'X': [0.5, 0.0, 0.5], 'W': [0.5, 0.25, 0.75],
              'K': [0.375, 0.375, 0.75], r'$\Gamma$': [0, 0, 0]}


def plot_recip_pt_labels(ax, labels, point_array, path_array, label_height="lower", show_line=True):
    """"
    Method to add reciprocal point labels to the plot

    Parameters
    ----------
    ax : matplotlib.axes.Axes
       Axis with plotted dispersion

    labels : dict
       Dictionary with keys corresponding to reciprocal point labels as strings, and values
       corresponding to array_like reciprocal points

    point_array : array_like
       Array of points corresponding to the x-axis of the plot

    path_array : array_like
       The path coordinates corresponding to the point_array

    label_height : str
       Can either be "upper" or "lower" to specify if labels should be above or below the plot

    show_line : bool
       If true, a line will be plotted to mark labeled reciprocal points

    Returns
    -------
    ax: matplotlib.axes.Axes
       Axis with the plotted dispersion and labeled reciprocal points

    """

    if label_height == "upper":
        label_height = ax.get_ylim()[1] + (ax.get_ylim()[1] - ax.get_ylim()[0]) * 0.1

    elif label_height == "lower":
        label_height = ax.get_ylim()[0] - (ax.get_ylim()[1] - ax.get_ylim()[0]) * 0.1

    for label in labels.keys():
        path_to_label = lattice.convert_point2path(labels[label], point_array, path_array)

        if path_to_label is None:
            continue
        for x in path_to_label:
            if show_line:
                ax.axvline(x, c="lightgray")
                ax.text(x=x, y=label_height, s=label)

    return ax


def set_energy_window(ax, energy_window):
    """"
    Method to add reciprocal point labels to the plot

    Parameters
    ----------
    ax : matplotlib.axes.Axes
       Axis with plotted dispersion

    energy_window : tuple
       Tuple with the lower and upper bound of the chosen energy window

    Returns
    -------
    ax: matplotlib.axes.Axes
       Axis with the plotted dispersion and adjusted energy window

    """
    ax.set_ylim((float(energy_window[0]) * 1.01, float(energy_window[1]) * .99))
    return ax


def plot_dispersion(ax, path, energies, energy_units, c="k", ls='-', energy_window=None):
    """
    Method to plot the dispersion (phonon dispersion or band structure).

    Parameters
    ----------
    ax: matplotlib.axes.Axes
       Axis on which to plot the dispersion

    path : array_like
       The array or list of path coordinates to be plotted on the x-axis

    energies : dict
       Dictionary of arrays of energies to plot, with keys labelling the band number or phonon mode.

    energy_units : str

    c : str, list
       Matplotlib color for plotting. If a list, different colors will be iterated through as each band is plotted.

    ls : str, list
       Matplotlib linestyle for plotting. If a list, different linestyles will be iterated through as each band is plotted.

    energy_window : tuple, optional
       Tuple with the lower and upper bound of the chosen energy window

    Returns
    -------
    ax: matplotlib.axes.Axes
       Axis with the plotted dispesion

    """
    if isinstance(c, str):
        c = [c]

    if isinstance(ls, str):
        ls = [ls]

    for n in energies.keys():
        x = path
        y = energies[n]
        ax.plot(x, y, c=c[n % len(c)], linestyle=ls[n % len(ls)])

    if energy_window is not None:
        ax = set_energy_window(ax, energy_window)

    ax.set_xticks([])
    ax.set_ylabel(f'Energy ({energy_units})')

    return ax


def plot_vals_on_bands(ax, path, energies, energy_units, values, cmap='RdBu', energy_window=None, max_val=None):
    """
    Method to plot the dispersion (phonon dispersion or band structure).

    Parameters
    ----------
    ax: matplotlib.axes.Axes
       Axis on which to plot the dispersion

    path : array_like
       The array or list of path coordinates to be plotted on the x-axis

    energies : dict
       Dictionary of arrays of energies to plot, with keys labelling the band number or phonon mode.

    energy_units : str

    values : dict
       Dictionary of arrays of values to plot on the bands by color

    cmap : str
       Matplotlib cmap for plotting values.

    energy_window : tuple, optional
       Tuple with the lower and upper bound of the chosen energy window

    Returns
    -------
    ax: matplotlib.axes.Axes
       Axis with the plotted dispesion

    """

    # Create a continuous norm to map from data points to colors
    vmin = min([min(values[key]) for key in values.keys()])
    vmax = max([max(values[key]) for key in values.keys()])

    if max_val is not None:
        vmax = max_val

    norm = plt.Normalize(vmin, vmax)

    for n in energies.keys():

        x = np.array(path)
        y = np.array(energies[n])

        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[: -1], points[1:]], axis=1)

        lc = LineCollection(segments, cmap=cmap, norm=norm)

        lc.set_array(values[n])
        lc.set_linewidth(2)
        line = ax.add_collection(lc)

        if ax.get_xlim()[0] > x.min():
            ax.set_xlim(ax.min(), ax.get_xlim()[1])
        if ax.get_xlim()[1] < x.max():
            ax.set_xlim(ax.get_xlim()[0], x.max())

        if ax.get_ylim()[0] > y.min():
            ax.set_ylim(y.min(), ax.get_ylim()[1])
        if ax.get_ylim()[1] < y.max():
            ax.set_ylim(ax.get_ylim()[0], y.max())

    plt.colorbar(line, ax=ax)

    if energy_window is not None:
        ax = set_energy_window(ax, energy_window)

    ax.set_xticks([])
    ax.set_ylabel(f'Energy ({energy_units})')

    return ax



import numpy as np
from scipy.optimize import curve_fit
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.postproc.utils.constants import energy_conversion_factor, length_conversion_factor
from perturbopy.postproc.dbs.units_dict import UnitsDict
from perturbopy.postproc.dbs.recip_pt_db import RecipPtDB
from perturbopy.postproc.utils.plot_tools import plot_dispersion, plot_recip_pt_labels
from perturbopy.postproc.utils.lattice import reshape_points, cryst2cart


class Bands(CalcMode):
    """
    Class representation of a Perturbo bands calculation.

    Parameters
    ----------
    pert_dict : dict
    Dictionary containing the inputs and outputs from the bands calculation.

    Attributes
    ----------
    kpt : RecipPtDB
       Database for the k-points used in the bands calculation.
    bands : UnitsDict
       Database for the band energies computed by the bands calculation.

    """

    def __init__(self, pert_dict):
        """
        Constructor method

        """
        super().__init__(pert_dict)

        if self.calc_mode != 'bands':
            raise ValueError('Calculation mode for a BandsCalcMode object should be "bands"')

        kpath_units = self._pert_dict['bands'].pop('k-path coordinate units')
        kpath = np.array(self._pert_dict['bands'].pop('k-path coordinates'))
        kpoint_units = self._pert_dict['bands'].pop('k-point coordinate units')
        kpoint = np.array(self._pert_dict['bands'].pop('k-point coordinates'))

        energies_dict = self._pert_dict['bands'].pop('band index')
        num_bands = self._pert_dict['bands'].pop('number of bands')
        energy_units = self._pert_dict['bands'].pop('band units')

        self.kpt = RecipPtDB.from_lattice(kpoint, kpoint_units, self.lat, self.recip_lat, kpath, kpath_units)
        self.bands = UnitsDict.from_dict(energies_dict, energy_units)

    def indirect_bandgap(self, n_lower, n_upper):
        """
        Method to compute the indirect bandgap between two bands.

        Parameters
        ----------
        n_lower, n_upper : int
           Band number of the lower and upper bands.

        Returns
        -------
        gap: float
           The indirect bandgap, computed as the energy difference between the minimum of
           the upper band and the maximum of the lower band.
        lower_kpoint, upper_kpoint : array
           k-points corresponding to the minimum of the upper band and the maximum of the lower band.

        Raises
        ------
        ValueError
           If the upper and lower band numbers provided are not valid band indices from the bands database, or
           if n_lower is greater than n_upper.

        """

        if n_lower not in self.bands.keys() or n_upper not in self.bands.keys():
            raise ValueError("n_lower and n_upper must be valid band numbers.")

        if n_lower > n_upper:
            raise ValueError("n_lower must be less than or equal to n_upper.")

        gap = np.min(self.bands[n_upper]) - np.max(self.bands[n_lower])

        lower_kpoint = self.kpt.points[:, np.argmax(self.bands[n_lower])]
        upper_kpoint = self.kpt.points[:, np.argmin(self.bands[n_upper])]

        return gap, lower_kpoint, upper_kpoint

    def direct_bandgap(self, n_lower, n_upper):
        """
        Method to compute the direct bandgap between two bands.

        Parameters
        ----------
        n_lower, n_upper : int
           Band number of the lower and upper bands.

        Returns
        -------
        gap: float
           The direct bandgap, computed as the minimum energy difference between two bands
           at the same k-point.
        kpoint: array
           The k-point corresponding to the direct bandgap.

        Raises
        ------
        ValueError
           If the upper and lower band numbers provided are not valid band indices from the bands database, or
           if n_lower is greater than n_upper.

        """
        if n_lower not in self.bands.keys() or n_upper not in self.bands.keys():
            raise ValueError("n_lower and n_upper must be valid band numbers")

        if n_lower > n_upper:
            raise ValueError("n_lower must be less than or equal to n_upper.")

        transitions = self.bands[n_upper] - self.bands[n_lower]
        gap = np.min(transitions)
        kpoint = self.kpt.points[:, np.argmin(transitions)]

        return gap, kpoint

    def effective_mass(self, n, kpoint, max_distance, direction=None, ax=None, c='r'):
        """
        Method to compute the effective mass at a k-point, approximated with a parabolic fit.

        Parameters
        ----------
        n : int
           Index of the band for which to calculate the effective mass.

        kpoint : list
           The k-point on which to center the calculation.

        max_distance : float
           Maximum distance between the center k-point and k-points to include in the parabolic approximation.

        direction : array_like, optional
           The k-point specifying the direction of the effective mass. Defaults to the same value as kpoint,
           i.e. the longitudinal effective mass.

        ax : matplotlib.axes.Axes
           Axis on which to plot the bands and approximated parabolic curve

        c : str
           Color for plotting the approximated parabolic curve

        Returns
        -------
        effective_mass : float
           The longitudinal effective mass at band n and the inputted kpoint, computed by a parabolic fit.

        """

        # Default direction is longitudinal, i.e. in same direction as central k-point
        if direction is None:
            direction = kpoint
        else:
            direction = reshape_points(direction)

        kpoint = reshape_points(kpoint)

        energies = self.bands[n] * energy_conversion_factor(self.bands.units, 'hartree')
        alat = self.alat * length_conversion_factor(self.alat_units, 'bohr')
        E_0 = energies[self.kpt.find(kpoint)][0]

        def get_fit_data(max_fit_distance, kpoint, direction, max_points=None, epsilon=1e-6):

            kpoint_distances = np.linalg.norm(self.kpt.points - np.array(kpoint), axis=0)
            kpoint_mag_squared = np.linalg.norm(self.kpt.points, axis=0)

            # Find all k-points parallel to the direction within a tolerance, epsilon
            kpoint_parallel = abs(np.divide(np.dot(np.reshape(direction, (3,)), self.kpt.points), (np.linalg.norm(direction) * kpoint_mag_squared),
                                            where=kpoint_mag_squared != 0) - 1) < epsilon

            if max_points is None:
                kpoint_indices = np.where(np.logical_and(kpoint_distances < max_fit_distance, kpoint_parallel))[0]

                kpoint_idx = self.kpt.find(kpoint)[0]

                if kpoint_idx not in kpoint_indices.flatten():
                    kpoint_indices = np.append(kpoint_indices, kpoint_idx)

                kpoint_indices = np.sort(kpoint_indices)
            else:
                kpoint_indices = np.where(kpoint_parallel)[0]
                kpoint_idx = self.kpt.find(kpoint)[0]
                kpoint_indices = np.where(abs(kpoint_indices - kpoint_idx) <= max_points)[0]

            kpoint_idx = self.kpt.find(kpoint)[0]

            kpt_points = self.kpt.points[:, kpoint_indices]

            kpt_points = cryst2cart(kpt_points, self.lat, self.recip_lat, forward=True, real_space=False)
            kpoint = cryst2cart(kpoint, self.lat, self.recip_lat, forward=True, real_space=False)

            kpoint_distances_squared = np.sum(np.square(kpt_points - kpoint), axis=0) * (np.pi * 2 / self.alat) ** 2

            return kpoint_indices, kpoint_distances_squared

        def parabolic_approx(kpoint_dist_squared, prefactor):
            return prefactor * kpoint_dist_squared + E_0

        fit_indices, fit_distances_squared = get_fit_data(max_distance, kpoint, direction)
        fit_energies = energies[fit_indices]
        fit_params, pcov = curve_fit(parabolic_approx, fit_distances_squared, fit_energies)

        effective_mass = 1 / (fit_params[0] * 2)

        if ax is not None:
            ax = self.plot_bands(ax, 'k')

            plot_indices, plot_distances_squared = get_fit_data(max_distance * 1.8, kpoint, direction)
            energies_fitted = (fit_params[0] * plot_distances_squared + E_0) * energy_conversion_factor('hartree', self.bands.units)

            ax.plot(self.kpt.path[plot_indices], energies_fitted, c, marker=None, ls='--')
            ax.plot(self.kpt.path[fit_indices], energies[fit_indices] * energy_conversion_factor('hartree', self.bands.units), c, marker='o')

        return effective_mass

    def plot_bands(self, ax, show_kpoint_labels=True, c='k', ls='-', energy_window=None):
        """
        Method to plot the band structure.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
           Axis on which to plot the bands.

        energy_window : tuple of int, optional
           The range of band energies to be shown on the y-axis.

        show_kpoint_labels : bool, optional
           If true, the k-point labels stored in the labels attribute will be shown on the plot. Default true.

        Returns
        -------
        ax: matplotlib.axes.Axes
           Axis with the plotted bands.

        """
        ax = plot_dispersion(ax, self.kpt.path, self.bands, self.bands.units, c, ls, energy_window)

        if show_kpoint_labels:
            ax = plot_recip_pt_labels(ax, self.kpt.labels, self.kpt.points, self.kpt.path)

        return ax



import numpy as np
from perturbopy.postproc.utils.constants import standardize_units_name, recip_points_units_names
from perturbopy.postproc.utils.plot_tools import points_fcc
from perturbopy.postproc.utils import lattice


class RecipPtDB():
    """
    This is a class representation of a set of points in reciprocal space.

    Attributes
    ----------
    points_cart : array_like
       Array of reciprocal space points in cartesian coordinates, in units of 2pi/a.

    points_cryst : array_like
       Array of reciprocal space points in crystal coordinates.

    units : str
       The units that calculations will be performed in (crystal or cartesian).
       If units = 'crystal', then points_cryst will be used instead of points_cart when
       a method such as "distances" is called on the RecipPtDB object.

    points : array_like
       Array of reciprocal space points in crystal or cartesian coordinates, depending on the
       value of "units".
       
    path : array_like
       Array of floats corresponding to each reciprocal space point, for plotting purposes.

    path_units : str
       Units of path, typically arbitrary

    labels : dict
       Dictionary of reciprocal space point labels
       example: {"Gamma": [0, 0, 0], 'L': [.5,.5,.5]}

    """

    def __init__(self, points_cart, points_cryst, units='crystal', path=None, path_units='arbitrary', labels={}):

        """
        Constructor method

        Parametrs
        ---------
        points_cart : array_like
           Array of reciprocal space points in cartesian coordinates, in units of 2pi/a.

        points_cryst : array_like
           Array of reciprocal space points in crystal coordinates.

        units : str
           The units that calculations will be performed in (crystal or cartesian).
           If units = 'crystal', then points_cryst will be used instead of points_cart when
           a method such as "distances" is called on the RecipPtDB object.

        path : array_like
           Array of floats corresponding to each reciprocal space point, for plotting purposes.

        path_units : str
           Units of path, typically arbitrary

        labels : dict
           Dictionary of reciprocal space point labels
           example: {"Gamma": [0, 0, 0], 'L': [.5,.5,.5]}
        """
        self.points_cart = lattice.reshape_points(points_cart)
        self.points_cryst = lattice.reshape_points(points_cryst)
        self.units = standardize_units_name(units, recip_points_units_names)

        if self.units == 'cartesian':
            self.points = self.points_cart

        elif self.units == 'crystal':
            self.points = self.points_cryst

        if path is None:
            self.path = np.arange(0, np.shape(self.points_cart)[1])
        else:
            self.path = path

        self.path_units = path_units

        self.labels = labels.copy()

    @classmethod
    def from_lattice(self, points, units, lat, recip_lat, path=None, path_units='arbitrary', labels={}):
        """
        Class method to create a RecipPtDB from one set of reciprocal space points and the lattice information.

        Parameters
        ----------
        points : array_like
           Array of column-oriented reciprocal space points in cartesian or crystal coordinates. If clearly not
           column-oriented (i.e. the number of columns, but not rows, is 3), points will be reshaped.

        units : str
           The units points are given in

        lat : array
           3x3 array of lattice vectors [v1, v2, v3] in units of alat

        recip_lat : array
           3x3 array of reciprocal lattice vectors [v1, v2, v3] in units of
           2pi/a

        path : array, optional
           Array of values corresponding to each reciprocal space point, for plotting purposes

        path_units : str, optional
           Units of the path points, typically arbitrary

        labels : dict, optional
           Dictionary of reciprocal space point labels
           example: {"Gamma": [0, 0, 0], 'L': [.5,.5,.5]}

        Returns
        -------
        points_db : RecipPtDB
           The RecipPtDB created from the lattice information and reciprocal space points

        """
        units = standardize_units_name(units, recip_points_units_names)

        if units == 'cartesian':
            points_cart = lattice.reshape_points(points)
            points_cryst = lattice.cryst2cart(points_cart, lat, recip_lat, forward=False, real_space=False)

        elif units == 'crystal':

            points_cryst = lattice.reshape_points(points)
            points_cart = lattice.cryst2cart(points_cryst, lat, recip_lat, forward=True, real_space=False)
        else:
            return None

        return RecipPtDB(points_cart, points_cryst, units, path, path_units, labels)

    def convert_units(self, new_units, in_place=True):
        """
        Method to convert the assumed units between crystal and cartesian coordinates.

        new_units : str
           The new units to use, either crystal or cartesian (or some valid variation of those names)

        in_place : bool, optional
           If true, the new units will be assumed in all future calls to the RecipPtDB.

        Returns
        -------
        points : array
           The stored points in the new units

        """

        self.units = standardize_units_name(new_units, recip_points_units_names)

        if self.units == 'cartesian':
            self.points = self.points_cart

        elif self.units == 'crystal':
            self.points = self.points_cryst

        return self.points

    def scale_path(self, range_min, range_max):
        """
        Method to scale the arbitrary k path plotting coordinates to a certain range.

        Parameters
        ----------
        range_min : float
           Lower limit of the range to which the k path coordinates will be scaled.

        range_max : float
           Upper limit of the range to which the k path coordinates will be scaled.

        Returns
        -------
        None

        """

        self.path = (self.path - min(self.path)) \
                                 / (max(self.path) - min(self.path)) \
                                 * (range_max - range_min) \
                                 + range_min

    def distances(self, point):
        """
        Method to compute the distances between each reciprocal space point in the points property
        and an inputted reciprocal space point

        Parameters
        ----------
        point : array
           The reciprocal space point that distances will be computed from

        Returns
        -------
        distances : array
           an array of distances between each reciprocal space point in the database and the inputted point

        """
        distances = lattice.compute_distances(self.points, point)

        return distances

    def find(self, point, max_dist=0.025, nearest=True):
        """
        Method to find the index or indices of a particular point

        Parameters
        ----------
        point : array
           The reciprocal space point to be searched

        max_dist : float, optional
           The maximum distance between the point to locate and the points identified as matches

        nearest : bool, optional
           If True, only the nearest match, or matches in the case of duplicate points, are returned (even if
           other points are within the max_dist)

        Returns
        -------
        points_indices : list
           The indices of the matching reciprocal space point in the points array

        """
        points_indices = lattice.find_point(point, self.points, max_dist, nearest)

        return points_indices

    def point2path(self, point, max_dist=0.025, nearest=True):
        """
        Method to find the path coordinate corresponding to a reciprocal space point coordinate

        Parameters
        ----------
        point : array_like
           The reciprocal space point to be searched

        max_dist : float, optional
           The maximum distance between the point to locate and the points identified as matches

        nearest : bool, optional
           If True, only the nearest match, or matches in the case of duplicate points, are returned (even if
           other points are within the max_dist)

        Returns
        -------
        path_coord : array
           The path coordinates of the corresponding reciprocal space point(s)

        """

        path_coord = lattice.convert_point2path(point, self.points, self.path, max_dist, nearest)

        return path_coord

    def path2point(self, path_coord, atol=1e-5, rtol=1e-2, nearest=True):
        """
        Method to find the reciprocal space point corresponding to a path coordinate

        Parameters
        ----------
        path_coord : float
           The path coordinate to be converted to a point

        atol : float, optional
        The absolute tolerance between the path coordinate to locate and the matching path coordinates

        rtol : float, optional
           The relative tolerance between the path coordinate to locate and the matching path coordinates

        nearest : bool, optional
           If True, only the nearest path coordinate, or coordinates in the case of repeats, are returned (even if
           other points are within the absolute and relative tolerance)

        Returns
        -------
        point : array
           The reciprocal space point(s) of the corresponding path coordinate

        """

        point = lattice.convert_path2point(path_coord, self.points, self.path, atol, rtol, nearest)

        return point

    def add_labels(self, labels_dict_input):
        """
        Method to add labels associated with a reciprocal space point. For example, point = [0,0,0] and label = 'gamma'

        Parameters
        ----------
        points : list
           A list of reciprocal space points

        labels_dict_input : dict
           A dictionary with the keys corresponding to reciprocal space point names, and values corresponding tolerance
           reciprocal space points

        """
        labels_dict = labels_dict_input.copy()
        for i, label in enumerate(labels_dict.keys()):
            self.labels[label] = labels_dict[label]

    def remove_labels(self, labels_list):
        """
        Method to remove a label associated with a reciprocal space point. For example, point = [0,0,0] and label = 'gamma'

        Parameters
        ----------
        labels_list : array_like
           The set of labels (keys in self.labels) to remove

        """

        for label in labels_list:
            self.labels.pop(label)



"""
How this works.

In the tests fodler, the test_perturbo.py script runs Perturbo for
a given 'test_name'. The idea is to use the same function (test_perturbo())
to run all of the calc. modes, epr files, etc. All of this should be
happening under the pytest environment (ran by `pytest`).

In this file (conftest.py), we parametrize pytest to make this work.
"""

from perturbopy.test_utils.run_test.run_utils import get_all_tests
from perturbopy.test_utils.run_test.run_utils import filter_tests
from perturbopy.test_utils.run_test.test_driver import clean_epr_folders
import pytest
import os

# define all supplementary arguments for the test running. This function declared in the PyTest itself


def pytest_addoption(parser):

    parser.addoption('--source_folder',
                     help='address of the folder with all reference files for the test performance',
                     nargs="?",
                     default='./')

    parser.addoption('--tags',
                     help='List of tests tags to include in this testsuite run.',
                     nargs='*', default=None)

    parser.addoption('--exclude-tags',
                     help='List of tests tags to exclude from this testsuite run.',
                     nargs='*', default=None)
                     
    parser.addoption('--epr_tags',
                     help='List of epr_tags to include in this testsuite run.',
                     nargs='*', default=None)

    parser.addoption('--exclude-epr_tags',
                     help='List of epr_tags to exclude from this testsuite run.',
                     nargs='*', default=None)

    parser.addoption('--epr',
                     help='List of epr files to test.',
                     nargs='*', default=None)

    parser.addoption('--test-names',
                     help='List of test names to include in this testsuite run.',
                     nargs='*', default=None)

    parser.addoption('--devel',
                     help='Include the development-stage tests.',
                     action='store_true', default=False)

    parser.addoption('--run_qe2pert',
                     help='Include the qe2pert tests',
                     action='store_true')
                     
    parser.addoption('--config_machine',
                     help='Name of file with computational information for qe2pert computation. Should be in the folder tests_f90/comp_qe2pert',
                     nargs="?", default='config_machine.yml')
                     
    parser.addoption('--keep_perturbo',
                     help='Save all the materials related to perturbo tests',
                     action='store_true')
                     
    parser.addoption('--keep_epr',
                     help='Save all epr-files from the qe2pert testing',
                     action='store_true')
                     
    parser.addoption('--keep_preliminary',
                     help='Save all preliminary files for epr files calculations',
                     action='store_true')

# generation of test for each type of test function. This function automatically called,
# when we run tests from the subfolders of folder, where this function is saved.


def pytest_generate_tests(metafunc):
    """
    The purpose of this function is to feed multiple test names to the
    tests/test_perturbo.py test_perturbo(test_name) function.
    Here, this function is referred as 'metafunc'.

    First, we get the list of all of the test folders using the tests/epr_info.yml file.
    We retrieve the test folder as <epr name>-<test name>. This is done by the get_all_tests()
    function.

    Next, we remove some of the tests based on the command line options and obtain the target
    test_list using the filer_tests() function.

    Finally, we feed the elements of the test_list to the test_perturbo() (metafunction)
    as the test_name input argument of the function.
    """
    if 'test_name' in metafunc.fixturenames:
        
        # define folder with supplementary information
        source_folder = os.path.abspath(metafunc.config.getoption('source_folder'))
        
        if not os.path.exists((os.path.join(source_folder, f'config_machine/{metafunc.config.getoption("config_machine")}'))):
            raise FileNotFoundError(f"File {metafunc.config.getoption('config_machine')} not found in the {source_folder}/config_machine folder. "
                                    f"Please create this file. In the folder {source_folder}/config_machine you can find examples of configurational files. "
                                    "Make your own copy with the name config_machine.yml in the same folder and run the testsuite again.")

        # Get the list of all test folders
        all_test_list, all_dev_test_list = get_all_tests(metafunc.function.__name__, source_folder)

        # Add the devel tests if --devel was specified
        if metafunc.config.getoption('devel'):
            all_test_list += all_dev_test_list

        if (metafunc.function.__name__ == 'test_perturbo') or (metafunc.function.__name__ == 'test_perturbo_for_qe2pert'):
            # sort out test folders based on command-line options (if present)
            test_list = filter_tests(
                all_test_list,
                metafunc.config.getoption('tags'),
                metafunc.config.getoption('exclude_tags'),
                metafunc.config.getoption('epr'),
                metafunc.config.getoption('test_names'),
                metafunc.function.__name__,
                metafunc.config.getoption('run_qe2pert'),
                source_folder
            )
        elif (metafunc.function.__name__ == 'test_qe2pert'):
            test_list = filter_tests(
                all_test_list,
                metafunc.config.getoption('epr_tags'),
                metafunc.config.getoption('exclude_epr_tags'),
                metafunc.config.getoption('epr'),
                None,
                metafunc.function.__name__,
                metafunc.config.getoption('run_qe2pert'),
                source_folder
            )
        
        if metafunc.function.__name__ == 'test_perturbo_for_qe2pert':
            metafunc.parametrize('test_name', test_list, indirect=True)
            metafunc.parametrize('config_machine', [metafunc.config.getoption('config_machine')], indirect=True)
            metafunc.parametrize('keep_perturbo', [metafunc.config.getoption('keep_perturbo')], indirect=True)
            metafunc.parametrize('run_qe2pert', [metafunc.config.getoption('run_qe2pert')], indirect=True)
            metafunc.parametrize('source_folder', [source_folder], indirect=True)
        elif metafunc.function.__name__ == 'test_qe2pert':
            metafunc.parametrize('test_name', test_list, indirect=True)
            metafunc.parametrize('run_qe2pert', [metafunc.config.getoption('run_qe2pert')], indirect=True)
            metafunc.parametrize('config_machine', [metafunc.config.getoption('config_machine')], indirect=True)
            metafunc.parametrize('source_folder', [source_folder], indirect=True)
        elif metafunc.function.__name__ == 'test_perturbo':
            metafunc.parametrize('test_name', test_list, indirect=True)
            metafunc.parametrize('config_machine', [metafunc.config.getoption('config_machine')], indirect=True)
            metafunc.parametrize('keep_perturbo', [metafunc.config.getoption('keep_perturbo')], indirect=True)
            metafunc.parametrize('source_folder', [source_folder], indirect=True)
        
# in order to properly run pytest functions, we need to declare fixtures,
# which allow us to use parametrization of test functions


@pytest.fixture
def test_name(request):
    return request.param
    

@pytest.fixture
def run_qe2pert(request):
    return request.param


@pytest.fixture
def config_machine(request):
    return request.param
    

@pytest.fixture
def keep_perturbo(request):
    return request.param


@pytest.fixture
def source_folder(request):
    return request.param

# this is predefined function of PyTest, which is runned after the end of all tests.
# here we delete all epr-folders, for which all tests were passed


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    if config.getoption('run_qe2pert'):
        if exitstatus == 0:
            # delete all epr files
            epr_failed = []
        else:
            failed_reports = terminalreporter.getreports('failed')
            # Process the list of failed test reports as needed
            epr_failed = set()
            for report in failed_reports:
                # obtain epr-name of failed test
                epr_failed.add(report.nodeid.split("[")[1].rstrip("]").split('-')[0])
            epr_failed = list(epr_failed)
    
        clean_epr_folders(epr_failed, config.getoption('config_machine'), config.getoption('keep_epr'),
                          config.getoption('keep_preliminary'), os.path.abspath(config.getoption('source_folder')))



"""
   This module is wrapper to functions that compare yaml or HDF5 files
"""
import perturbopy.test_utils.compare_data.yaml as cyaml
import perturbopy.test_utils.compare_data.h5 as ch5


def equal_values(file1, file2, ig_n_tol):
    """
    Compare two files:
       #. determines file types
       #. determins if files containe the same values for both yaml/h5 the objects of the highest level are checked then checks are performed on common objects

    Parameters
    ----------
    file1 : str
       first  filename

    file2 : str
       second filename
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on files

    Returns
    -------
    equal_vlaues : bool
       boolean specifying if both files contain the same group/datasets/keys/values

    """
    # get file extension
    file1_type = file1.split(".")[-1]
    file2_type = file2.split(".")[-1]

    errmsg = ("file1 and file2 extension do not match"
              "file1_type = {file1_type}\n"
              "file2_type = {file2_type}\n")
    assert file1_type == file2_type, errmsg

    file_type = file1_type

    # use appropriate compare function for file type
    if file_type == "yaml" or file_type == "yml":
        equal_values, diff = cyaml.equal_values(file1, file2, ig_n_tol)

    elif file_type == "h5" or file_type == "hdf5":
        equal_values, diff = ch5.equal_values(file1, file2, ig_n_tol)

    else:
        errmsg = ("Can only compare files with extensions "
                  "yaml, yml, hdf5, or h5")
        raise ValueError(errmsg)

    if not equal_values:
        print(f' difference: {diff}')

    return equal_values



import numpy as np
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.postproc.dbs.units_dict import UnitsDict
from perturbopy.postproc.dbs.recip_pt_db import RecipPtDB
from perturbopy.postproc.utils.constants import hbar


class Imsigma(CalcMode):
    """
    Class representation of a Perturbo imsigma calculation.

    Attributes
    ----------
    kpt : RecipPtDB
       Database for the k-points used in the imsigma calculation, containing N points.
    
    bands : UnitsDict
       Database for the band energies computed by the imsigma calculation. The keys are
       the band index, and the values are an array (of length N) containing the energies at each k-point
       with units bands.units

    temper : UnitsDict
        Dictionary of temperatures used in each configuration. The keys give the configuration number,
        and the values are floats giving the temperature (with units temper.units)
    
    chem_pot : UnitsDict
        Dictionary of chemical potentials used in each configuration. The keys
        give the configuration number, and the values are floats giving the
        chemical potential (with units chem_pot.units)
    
    imsigma : UnitsDict
        Dictionary of imaginary self-energies computed for each configuration. The top level keys are the
        configuration number, and the second level keys are the band index. The values are arrays of length N giving the
        imaginary self-energies along all the k-points at that band index for the configuration. Units are in imsigma.units.

    imsigma_mode : UnitsDict
        Dictionary of imaginary self-energies resolved by phonon mode computed. The top level keys are the
        configuration number, and the second level keys are the band index. The third level keys are
        the phonon mode. Finally,the values are arrays of length N giving the imaginary self-energies along all the k-points
        due to the given phonon mode at that band index for the configuration. Units are in imsigma_mode.units.
    
    """

    def __init__(self, pert_dict):
        """
        Constructor method

        Parameters
        ----------
        pert_dict : dict
            Dictionary containing the inputs and outputs from the imsigma calculation.

        """
        super().__init__(pert_dict)

        if self.calc_mode.split('-')[0] != 'imsigma':
            raise ValueError('Calculation mode for an ImsigmaCalcMode object should be "imsigma"')

        kpoint_units = self._pert_dict['imsigma'].pop('k-point coordinate units')
        num_kpoints = self._pert_dict['imsigma'].pop('number of k-points')
        kpoint = np.array(self._pert_dict['imsigma'].pop('k-point coordinates'))
        self.kpt = RecipPtDB.from_lattice(kpoint, kpoint_units, self.lat, self.recip_lat)
        
        energy_units = self._pert_dict['imsigma'].pop('energy units')
        num_bands = self._pert_dict['imsigma'].pop('number of bands')
        energies_dict = self._pert_dict['imsigma']['energy'].pop('band index')
        self.bands = UnitsDict.from_dict(energies_dict, energy_units)

        num_config = self._pert_dict['imsigma'].pop('number of configurations')
        config_dat = self._pert_dict['imsigma'].pop('configuration index')
        num_modes = self._pert_dict['imsigma'].pop('number of phonon modes')

        self.temper = UnitsDict(units=self._pert_dict['imsigma'].pop('temperature units'))
        self.chem_pot = UnitsDict(units=self._pert_dict['imsigma'].pop('chemical potential units'))
        self.imsigma = UnitsDict(units=self._pert_dict['imsigma'].pop('Im(Sigma) units'))
        self.imsigma_mode = UnitsDict(units=self.imsigma.units)

        for config_idx in config_dat.keys():
            
            self.imsigma[config_idx] = {}
            self.imsigma_mode[config_idx] = {}
            self.temper[config_idx] = config_dat[config_idx].pop('temperature')
            self.chem_pot[config_idx] = config_dat[config_idx].pop('chemical potential')
            
            imsigma_dat = config_dat[config_idx].pop('band index')

            for mode in np.arange(1, num_modes + 1):
                self.imsigma_mode[config_idx][mode] = {}

            for band_index in imsigma_dat.keys():
                
                self.imsigma[config_idx][band_index] = np.array(imsigma_dat[band_index]['Im(Sigma)']['total'])

                for mode in np.arange(1, num_modes + 1):
                    self.imsigma_mode[config_idx][mode][band_index] = np.array(imsigma_dat[band_index]['Im(Sigma)']['phonon mode'][mode])



import numpy as np
import pytest
import os

import perturbopy.postproc as ppy


@pytest.fixture()
def gaas_bands():
    """
    Method to generate the BandsCalcMode object corresponding to GaAs YAML file.

    Returns
    -------
    bands : ppy.BandsCalcMode

    """
    yml_path = os.path.join("refs", "gaas_bands.yml")
    return ppy.Bands.from_yaml(yml_path)


@pytest.mark.parametrize("n_lower, n_upper, expected_gap, expected_kpoint", [
                         [4, 5, 0.29572179519999864, [0.01442, 0.01442, 0.02885]],
                         [8, 9, 0.44878296240000104, [0., 0., 0.]]
])
def test_direct_bandgap(gaas_bands, n_lower, n_upper, expected_gap, expected_kpoint):
    """
    Method to test direct_bandgap

    Parameters
    ----------
    n_lower, n_upper : int
       Band number of the lower and upper bands.

    expected_gap : float

    expected_kpoint : array_like

    """
    gap, kpoint = gaas_bands.direct_bandgap(n_lower, n_upper)
    assert(np.isclose(expected_gap, gap))
    assert(np.allclose(expected_kpoint, kpoint))

@pytest.mark.parametrize("n_lower, n_upper, expected_gap, expected_kpoint_lower, expected_kpoint_upper", [
                         [8, 9, 0.44878296240000104, [0., 0., 0.], [0., 0., 0.]],
                         [10, 11, -0.2976607195999996, [0.21635, 0.21635, 0.43269], [0.5, 0., 0.5]]
])
def test_indirect_bandgap(gaas_bands, n_lower, n_upper, expected_gap, expected_kpoint_lower, expected_kpoint_upper):
    """
    Method to test indirect_bandgap

    Parameters
    ----------
    n_lower, n_upper : int
       Band number of the lower and upper bands.

    expected_gap : float

    expected_kpoint_lower, expected_kpoint_upper : array_like

    """
    gap, kpoint_lower, kpoint_upper = gaas_bands.indirect_bandgap(n_lower, n_upper)
    assert(np.isclose(expected_gap, gap))
    assert(np.allclose(expected_kpoint_lower, kpoint_lower))
    assert(np.allclose(expected_kpoint_upper, kpoint_upper))


@pytest.mark.parametrize("n, kpoint, max_distance, direction, expected_m", [
                         [9, [0, 0, 0], 0.1, [0.5, 0.5, 0.5], 0.11826539850425859]
])
def test_effective_mass(gaas_bands, n, kpoint, max_distance, direction, expected_m):
    """
    Method to test effective_mass

    Parameters
    ----------
    n : int
       Index of the band for which to calculate the effective mass.

    kpoint : list
       The k-point on which to center the calculation.

    max_distance : float
       Maximum distance between the center k-point and k-points to include in the parabolic approximation.

    direction : array_like
       The k-point specifying the direction of the effective mass. Defaults to the same value as kpoint,
       i.e. the longitudinal effective mass.

    expected_m : float
       The expected effective mass computed

    """
    print(gaas_bands.bands)
    m = gaas_bands.effective_mass(n, kpoint, max_distance, direction)
    assert(np.isclose(expected_m, m))



import pytest
import math

import perturbopy.postproc as ppy

test_dict = {'bohr': ['bohr', 'a.u', 'atomic units', 'au'], 'angstrom': ['angstrom, a'], 'm': ['m', 'meter']}
test_vals = {'bohr': (1, 0), 'angstrom': (0.529177249, 0), 'm': (5.29177249, -11)}


@pytest.mark.parametrize("test_prefix, expected_exponent", [
                        ('m', -3), ('M', 6), ('mu', -6), ('Y', 24), ('', 0)
])
def test_prefix_exp(test_prefix, expected_exponent):
    """
    Test the constants.prefix_exp function

    Parameters
    ----------
    test_prefix : str
       The prefix to be converted to an exponent
    expected_exponent : int
       The expected exponent that the function should return

    """
    assert(ppy.constants.prefix_exp(test_prefix) == expected_exponent)


@pytest.mark.parametrize("test_units, expected", [
                        ('a.u', ('', 'bohr')), ('A.U', ('', 'bohr')), ('Bohr', ('', 'bohr')),
                        ('nm', ('n', 'm'))
])
def test_find_prefix_and_base_units(test_units, expected):
    """
    Test the constants.find_prefix_and_base_units function

    Parameters
    ----------
    test_units : str
       The units to be decomposed into prefix and base
    expected : tuple
       The expected prefix and base units to be returned

    """
    assert(ppy.constants.find_prefix_and_base_units(test_units, test_dict) == expected)


@pytest.mark.parametrize("test_units, expected_units", [
                        ('a.u', 'bohr'), ('A.U', 'bohr'), ('Bohr', 'bohr'),
                        ('nm', 'nm')
])
def test_standardize_units_name(test_units, expected_units):
    """
    Test the constants.find_prefix_and_base_units function

    Parameters
    ----------
    test_units : str
       The units to be converted to a standard units name
    expected_units : str
       The expected standardized units to be returned

    """
    assert(ppy.constants.standardize_units_name(test_units, test_dict) == expected_units)


@pytest.mark.parametrize("test_units1, test_units2, expected_factor", [
                        ('a.u', 'bohr', 1), ('a.u', 'nm', 5.29177249e-2),
                        ('cm', 'fm', 1e13), ('fm', 'cm', 1e-13),
                        ('a.u', 'bohr', 1), ('a.u', 'nm', 5.29177249e-2),
                        ('cm', 'fm', 1e13), ('fm', 'cm', 1e-13)
])
def test_conversion_factor(test_units1, test_units2, expected_factor):
    """
    Test the constants.conversion_factor function

    Parameters
    ----------
    test_units1, test_units2 : str
       The units between which to find a conversion factor
    expected_factor : float
       The expected conversion factor

    """
    assert(math.isclose(ppy.constants.conversion_factor(test_units1, test_units2, test_dict, test_vals), expected_factor))


@pytest.mark.parametrize("test_units1, test_units2, expected_factor", [
                        ('a.u', 'bohr', 1), ('a.u', 'nm', 5.29177249e-2),
                        ('cm', 'fm', 1e13), ('fm', 'cm', 1e-13)
])
def test_length_conversion_factor(test_units1, test_units2, expected_factor):
    """
    Test the constants.length_conversion_factor function

    Parameters
    ----------
    test_units1, test_units2 : str
       The units between which to find a conversion factor
    expected_factor : float
       The expected conversion factor

    """
    assert(math.isclose(ppy.constants.length_conversion_factor(test_units1, test_units2), expected_factor))


@pytest.mark.parametrize("test_units1, test_units2, expected_factor", [
                        ('meV', 'Ha', 3.674930882447527e-05), ('Ry', 'Ha', 2.0)
])
def test_energy_conversion_factor(test_units1, test_units2, expected_factor):
    """
    Test the constants.energy_conversion_factor function

    Parameters
    ----------
    test_units1, test_units2 : str
       The units between which to find a conversion factor
    expected_factor : float
       The expected conversion factor

    """
    print(ppy.constants.energy_conversion_factor(test_units1, test_units2))
    assert(math.isclose(ppy.constants.energy_conversion_factor(test_units1, test_units2), expected_factor))


@pytest.mark.parametrize("test_units, expected_hbar", [
                        ('ev*fs', 0.6582119569), ('atomic', 1)
])
def test_hbar(test_units, expected_hbar):
    """
    Test the constants.conversion_factor function

    Parameters
    ----------
    test_units : str
       The units for hbar
    expected_hbar : float
       The expected value of hbar in the corresponding units

    """
    assert(math.isclose(ppy.constants.hbar(test_units), expected_hbar))


def test_errors():
    """
    Test errors generated by constants module.

    """

    with pytest.raises(ValueError):
        ppy.constants.prefix_exp('ss')

    with pytest.raises(ValueError):
        ppy.constants.find_prefix_and_base_units('meV', test_dict)

    with pytest.raises(ValueError):
        ppy.constants.conversion_factor('meV', 'au', test_dict, test_vals)

    with pytest.raises(ValueError):
        ppy.constants.hbar('evs')
        ppy.constants.hbar(22)



#!/usr/bin/env python3
import time
from collections import OrderedDict


class Timing:
    """
    Class to measure timing for a specific task.

    Parameters
    ----------
    tag : str
        The tag or label for the timing measurement.

    Attributes
    ----------
    tag : str
        The label for the timing measurement.
    t_start : float or None
        The start time of the timing measurement.
    t_end : float or None
        The end time of the timing measurement.
    t_delta : float or None
        The time difference between the start and end times.
    total_runtime : float
        The accumulated total runtime for multiple measurements.
    call_count : int
        The number of times the timing has been measured.
    level : int
        The level of importance (hierarchy) for the timing measurement.

    Methods
    -------
    start()
        Start the timing measurement.
    stop()
        Stop the timing measurement and calculate the time difference.
    __enter__()
        Enter the context for the timing measurement.
    __exit__(exc_type, exc_val, exc_tb)
        Exit the context for the timing measurement.
    __str__()
        Return a string representation of the timing measurement.

    Examples
    --------
    # Example 1: Timing a code block using 'with' statement
    >>> with Timing('test') as t:
    ...     # Code to be timed
    ...     time.sleep(1)
    ...
    >>> print(t)
    test (s): 1.000

    # Example 2: Manually timing a code block without 'with' statement
    >>> t = Timing('test')
    >>> t.start()
    >>> # Code to be timed
    >>> time.sleep(1)
    >>> t.stop()
    >>> print(t)
    test (s): 1.000

    # Example 3: Accumulating multiple timing measurements
    >>> t = Timing('test')
    >>> t.start()
    >>> # Code to be timed
    >>> time.sleep(1)
    >>> t.stop()
    >>> t.start()
    >>> # Code to be timed again
    >>> time.sleep(0.5)
    >>> t.stop()
    >>> print(t)
    test (s): 1.500
    """

    def __init__(self, tag, level=0):
        self.tag = tag
        self.t_start = None
        self.t_end = None
        self.t_delta = None
        self.total_runtime = 0.0
        self.call_count = 0
        self.level = level

    def start(self):
        self.t_start = time.perf_counter()

    def stop(self):
        self.t_end = time.perf_counter()
        self.t_delta = self.t_end - self.t_start
        self.total_runtime += self.t_delta
        self.call_count += 1

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __str__(self):

        tag_text = self.tag
        lev = self.level

        width = 30
        lmax = 4

        if lev > 0:
            prefix = f'{chr(0x25B7)}'
            indent = ' ' * 3 * (lmax - lev)
            ndash = width - len(indent + prefix + tag_text)
            dashes = '-' * ndash
            tag_text = f'{indent}{prefix}{dashes}{tag_text}'

        output = f'{tag_text:>{width}}  {self.total_runtime:<12.3f} {self.call_count:<10}'

        return output


class TimingGroup:
    """
    Class to group and manage multiple timing measurements.

    Attributes
    ----------
    timings : OrderedDict
        A dictionary of tag-timing pairs.

    Methods
    -------
    add(tag, level=0)
        Add a new timing measurement with the given tag.
    sort()
        Sort the timings based on the total runtime.

    Examples
    --------
    >>> timing_group = TimingGroup()
    >>>
    >>> with timing_group.add('SVD') as t_svd:
    ...     # Code to be timed for SVD
    ...     time.sleep(1)
    ...
    >>> with timing_group.add('DIAGO') as t_diago:
    ...     # Code to be timed for DIAGO
    ...     time.sleep(2)
    ...
    >>> print(t_svd)
        SVD (s): Total Runtime: 1.000, Call Count: 1
    >>>
    >>> timing_group.print_timings()
        SVD    (s): Total Runtime: 1.000, Call Count: 1
        DIAGO  (s): Total Runtime: 2.000, Call Count: 1
    """
    def __init__(self, name=None):
        # self.timings = {}
        self.name = name
        self.timings = OrderedDict()

    def add(self, tag, level=0):
        """
        Add timing to existing tag or create a new one.
        """
        if tag in self.timings:
            timing = self.timings[tag]
        else:
            timing = Timing(tag, level)
            self.timings[tag] = timing

        return timing

    def sort(self):
        self.timings = dict(sorted(self.timings.items(), key=lambda x: x[1].total_runtime, reverse=True))

    def __str__(self):

        # self.sort()
        width = 60

        output = ''
        output += f"{'='*width}\n"

        line = f"{'='*20} {'Timing Summary'.upper()}"
        if self.name is None:
            end = f' {"="*(width - len(line)-1)}'
        else:
            end = f': {self.name} {"="*(width - len(line)-3-len(self.name))}'

        output += f'{line}{end}\n'
        output += f"{'='*width}\n"
        output += f'{"Tag":>30}  {"Time (s)":<12} {"Calls":<10}\n'
        output += f"{'-'*width}\n"

        for tag, timing_obj in self.timings.items():
            output += f'{timing_obj}\n'

        output += f"{'='*width}\n\n"

        return output

    def to_dict(self):
        """
        Return a dictionary of timings.

        Returns
        -------
        dict
            A dictionary containing the tag-timing pairs.

        """

        timings_dict = {tag:
                            {'runtime': round(timing.total_runtime, 3), 'call_count': timing.call_count}
                            for tag, timing in self.timings.items()
                        }

        # timings_dict = {tag: timing.total_runtime for tag, timing in self.timings.items()}
        return timings_dict


def measure_runtime_and_calls(method):
    """
    Decorator function to measure the runtime and calls of a method.
    """

    def wrapper(self, *args, **kwargs):
        if not hasattr(self, 'timings'):
            # Create TimingGroup instance as an attribute
            self.timings = TimingGroup()
        timings = self.timings
        with timings.add(method.__name__, level=1) as t:
            result = method(self, *args, **kwargs)
        return result
    return wrapper



"""
Suite of Python scripts for the Perturbo code testing and postprocessing.
"""
from .calc_modes.calc_mode import CalcMode
from .calc_modes.bands import Bands
from .calc_modes.phdisp import Phdisp
from .calc_modes.ephmat import Ephmat
from .calc_modes.trans import Trans
from .calc_modes.imsigma import Imsigma
from .calc_modes.dyna_run import DynaRun
from .calc_modes.dyna_pp import DynaPP

from .dbs.units_dict import UnitsDict
from .dbs.recip_pt_db import RecipPtDB

from .utils import constants, plot_tools, lattice



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
setup(
    name='perturbopy',
    version='0.1.1',
    description="Suite of Python scripts for Perturbo testing and postprocessing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.7.12',
    include_package_data=True,
    install_requires=[
        'numpy>=1.21.4',
        'pytest',
        'pytest-order',
        'h5py',
        'PyYAML',
        'pytest-profiling',
        'matplotlib>=2.2.0',
        'scipy'
    ],
    extras_require={
        'interactive': ['jupyter', 'pytest-plots'],
    },
    packages=find_packages(
        where='./src'
    ),
    package_dir={"": "src"}
)



import numpy as np
import pytest

import perturbopy.postproc as ppy

points_cryst = [[0, 0, 0.5, 0.25, 0.25, 0.375], [0, 0.5, 0.5, 0.75, 0.625, 0.75], [0, 0.5, 0.5, 0.5, 0.625, 0.375]]
points_cart = [[0, 0, 0.5, 0.5, 0.25, 0.75], [0, 1, 0.5, 1, 1, 0.75], [0, 0, 0.5, 0, 0.25, 0.]]


@pytest.fixture()
def recip_dbs():
    """
    Fixture to generate a RecipPtDB object

    """
    return ppy.RecipPtDB(points_cart, points_cryst, units='cryst')


@pytest.mark.parametrize("units_input, expected_units, expected_points", [
                        ('cryst', 'crystal', points_cryst), ('Cartesian', 'cartesian', points_cart),
])
def test_units(recip_dbs, units_input, expected_units, expected_points):
    """
    Method to test the conversion between different units

    Parameters
    ----------
    units_input : str
       The units to convert the points to
    expected_units : str
       The expected units to be stored, according to standardized units names
    expected_points : array
       The points expected to be stored in recip_dbs after conversion

    """
    recip_dbs.convert_units(units_input)
    assert(recip_dbs.units == expected_units)
    assert(np.all(recip_dbs.points == expected_points))


@pytest.mark.parametrize("test_min, test_max, expected", [
                        (0, 1, [0, 0.2, 0.4, 0.6, 0.8, 1])
])
def test_scale_path(recip_dbs, test_min, test_max, expected):
    """
    Method to test the recip_pt_db.scale_path function

    Parameters
    ----------
    test_min, test_max : float
       The min and max value in the range the path should be scaled to
    expected : list
       The expected, rescaled path
    """
    recip_dbs.scale_path(test_min, test_max)
    print(type(recip_dbs))
    assert(np.all(recip_dbs.path == expected))


@pytest.mark.parametrize("point, expected", [
                        ([[0, 0, 0.5, 0.25, 0.25, 0.375], [0, 0.5, 0.5, 0.75, 0.625, 0.75], [0, 0.5, 0.5, 0.5, 0.625, 0.375]], [0, 0, 0, 0, 0, 0]),
])
def test_compute_distances(recip_dbs, point, expected):
    """
    Method to test the recip_pt_db.compute_distances function

    Parameters
    ----------
    point : array_like
       The point to compute the distances between (compared to all the points in recip_dbs)
    expected : list
       The expected distances between the points in recip_dbs and the inputted point
    """
    print(recip_dbs.distances(point))
    assert(np.all(np.isclose(recip_dbs.distances(point), expected)))


@pytest.mark.parametrize("test_point, max_dist, expected", [
                        ([0, 0, 0], None, 0), ([0.25, 0.625, 0.625], None, 4),
                        ([0.25, 0.625, 0.56], None, []), ([0.25, 0.625, 0.56], 0.1, 4),
                        ([0, 0.5, 0.5], None, [1]),
])
def test_find(recip_dbs, test_point, max_dist, expected):
    """
    Method to test the recip_pt_db.find function

    Parameters
    ----------
    test_point : array_like
       The point to search recip_dbs
    max_dist : float
       The maximum distance between the point to locate and the points identified as matches
    expected : list
       The expected indices to be found

    """
    print(recip_dbs.distances(test_point))
    print(recip_dbs.find(test_point))
    if max_dist is None:
        assert(np.all(recip_dbs.find(test_point) == expected))
    else:
        assert(np.all(recip_dbs.find(test_point, max_dist=max_dist) == expected))


@pytest.mark.parametrize("test_point, max_dist, expected", [
                        ([0, 0, 0], None, 0), ([0.25, 0.625, 0.625], None, 4),
                        ([0.25, 0.625, 0.56], None, None), ([0.25, 0.625, 0.56], 0.1, 4),
])
def test_point2path(recip_dbs, test_point, max_dist, expected):
    """
    Method to test the recip_pt_db.point2path function

    Parameters
    ----------
    test_point : array_like
       The point to search recip_dbs
    max_dist : float
       The maximum distance between the point to locate and the points identified as matches
    expected : float
       The expected path coordinate(s) to be found

    """
    if max_dist is None:
        assert(np.all(recip_dbs.point2path(test_point) == expected))
    else:
        assert(np.all(recip_dbs.point2path(test_point, max_dist=max_dist) == expected))


@pytest.mark.parametrize("test_path, atol, expected", [
                        (0, None, [0, 0, 0]), (4, None, [0.25, 0.625, 0.625]), (4.1, None, None),
                        (4.1, 0.1, [0.25, 0.625, 0.625])
])
def test_path2point(recip_dbs, test_path, atol, expected):
    """
    Method to test the recip_pt_db.path2point function

    Parameters
    ----------
    test_path : float
       The path coordinate to search recip_dbs
    atol : float
       The maximum difference between the path coordinate to locate and the coordinates identified as matches
    expected : array
       The expected point(s) to be found

    """
    if atol is None:
        assert(np.all(recip_dbs.path2point(test_path) == expected))
    else:
        assert(np.all(recip_dbs.path2point(test_path, atol=atol) == expected))


@pytest.mark.parametrize("test_path, atol, expected", [
                        (0, None, [0, 0, 0]), (4, None, [0.25, 1, 0.25]), (4.1, None, None),
                        (4.1, 0.1, [0.25, 1, 0.25])
])
def test_path2point_cart(recip_dbs, test_path, atol, expected):
    """
    Method to test the recip_pt_db.path2point function after converting units

    Parameters
    ----------
    test_path : float
       The path coordinate to search recip_dbs
    atol : float
       The maximum difference between the path coordinate to locate and the coordinates identified as matches
    expected : array
       The expected point(s) to be found

    """
    recip_dbs.convert_units("cart")
    if atol is None:
        assert(np.all(recip_dbs.path2point(test_path) == expected))
    else:
        assert(np.all(recip_dbs.path2point(test_path, atol=atol) == expected))
